#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

from .phasedbeam import PhasedArrayBeam
from .sstbeam import SSTbeam, SSTbeamHPX
from .bstbeam import BSTbeam, BSTbeamHPX

__all__ = ['PhasedArrayBeam', 'SSTbeam', 'BSTbeam', 'SSTbeamHPX', 'BSTbeamHPX']


#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to compute a phased-array beam
        by A. Loh
"""

import os
import sys
import numpy as np

from astropy import constants as const

from .antenna import AntennaModel

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['PhasedArrayBeam']

class PhasedArrayBeam():
    """ Class PhasedArrayBeam
        Parameters:
        - p (ndarray): antenna positions ([[x1, y1, z1], [x2, y2, z2], ..])
        - m (ndarray): antenna model
        - a (float): azimuth of the target location
        - e (float): elevation of the target location
    """
    def __init__(self, p=None, m=None, a=180, e=45):
        self.pos   = p
        self.model = m
        self.azim  = a
        self.elev  = e
        self.resol = 0.2 

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def azim(self):
        """ Azimuth in degrees
        """
        return self._azim
    @azim.setter
    def azim(self, a):
        if not isinstance(a, (float, int, np.float32, np.float64)):
            raise TypeError("\n\t=== Azimuth must be a number ===")
        elif (a < 0.) or (a > 360.):
            a %= 360.
        else:
            pass
        self._azim = a
        return

    @property
    def elev(self):
        """ Elevation in degrees
        """
        return self._elev
    @elev.setter
    def elev(self, e):
        if not isinstance(e, (float, int, np.float32, np.float64)):
            raise TypeError("\n\t=== Elevation must be a number ===")
        elif (e < 0.) or (e > 90.):
            raise ValueError("\n\t=== Elevation should be between 0 and 90 ===")
        else:
            pass
        self._elev = e
        return

    @property
    def pos(self):
        """ Antenna position array
        """
        return self._pos
    @pos.setter
    def pos(self, p):
        if p is None:
            return
        if not isinstance(p, (np.ndarray)):
            raise TypeError("\n\t=== Antenna position should be an array ===")
        elif len(p.shape) != 2:
            raise ValueError("\n\t=== Antenna position should be a 2D array ===")
        elif p.shape[1] != 3:
            raise ValueError("\n\t=== Antenna position array should be a (x, 3) shaped array ===")
        else:
            self._pos = p
        return

    @property
    def model(self):
        """ Antenna model
        """
        return self._model
    @model.setter
    def model(self, m):
        if m is None:
            return
        if not isinstance(m, (AntennaModel, list)):
             raise ValueError("\n\t=== model should be a AntennaModel object / list ===")
        else:
            self._model = m
            return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getBeam(self):
        """ Compute the beam of the phased array
            Parameter:
            power (bool): return the power (False: return eiphi)
        """
        if isinstance(self.model, list):
            wavevec = 2 * np.pi * self.model[0].antenna_freq * 1.e6 / const.c.value
        else:
            wavevec = 2 * np.pi * self.model.antenna_freq * 1.e6 / const.c.value

        # ------ Sky grid ------ #
        thetagrid, phigrid = np.radians(np.meshgrid(np.arange(0, 90, self.resol),
            np.arange(0, 360, self.resol) ))
        thetagrid = np.fliplr(thetagrid)
        _xp = np.cos(thetagrid) * np.cos(phigrid)
        _yp = np.cos(thetagrid) * np.sin(phigrid)
        _zp = np.sin(thetagrid)

        # ------ Phase delay between antennae ------ #
        theta = np.repeat( np.radians( self.elev ), self.pos.shape[0])
        # theta = np.repeat( np.radians( 90. - self.elev ), self.pos.shape[0])
        phi   = np.repeat( np.radians( self.azim ), self.pos.shape[0])
        ux = np.cos(phi) * np.cos(theta)
        uy = np.sin(phi) * np.cos(theta)
        uz = np.sin(theta)
        dphix = wavevec * self.pos[:, 0] * ux
        dphiy = wavevec * self.pos[:, 1] * uy
        dphi  = dphix + dphiy

        # ------ Phase reference ------ #
        phi0  = wavevec * ( self.pos[:, 0] * _xp[:, :, np.newaxis] 
            + self.pos[:, 1] * _yp[:, :, np.newaxis] 
            + self.pos[:, 2] * _zp[:, :, np.newaxis] ).astype(np.float32)
        phase = phi0[:, :, :] - dphi[np.newaxis, np.newaxis, :]

        # ------ e^(i Phi) ------ #
        eiphi = np.sum( np.exp(1j * phase), axis=2 )
        eiphi = np.flipud(eiphi.T)
        if isinstance(self.model, list):
            antgain = np.zeros( (eiphi.shape) )
            for i in range(len(self.model)):
                antgain += np.flipud( self.model[i].antenna_gain(np.linspace(0, 360, eiphi.shape[1]),
                    np.linspace(0, 90, eiphi.shape[0])) )
            #antgain /= antgain.max()
        else:
            antgain = self.model.antenna_gain(np.linspace(0, 360, eiphi.shape[1]),
                np.linspace(0, 90, eiphi.shape[0]))

        beam = eiphi * eiphi.conjugate() * antgain
        beam = np.real(beam) / np.real(beam).max()
        return np.flipud(beam)









#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

from .read import SST, BST, XST
from .read import __all__ as allread

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.5.0'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'

__all__ = [] 
__all__.extend(allread)
# __all__.extend(['beam', 'skymodel', 'astro'])


#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
"""

from nenupy.hpx import Anabeam, Digibeam, Skymodel, Tracking, Transit
import pylab as plt
import numpy as np

from nenupy import BST, SST


# # TEST TRACKING
# b = BST('/Users/aloh/Desktop/NenuFAR_Data/20181030_200800_BST.fits')
# b.select(freq=60)

# t=Tracking()
# t.from_bst(obs='/Users/aloh/Desktop/NenuFAR_Data/20181030_200800_BST.fits', dt=240, resol=0.2, freq=60)

# x = (t.simulation['time'] - t.simulation['time'][0]).sec / 60
# y = t.simulation['amp']

# scale = np.median(b.data['amp']) / np.median(y)
# plt.plot(x, 10 * np.log10(y))

# plt.plot( (b.data['time']-t.simulation['time'][0]).sec/60, 10 * np.log10(b.data['amp']/scale) )

# plt.xlabel('Time (min since {})'.format(t.simulation['time'][0].iso))
# plt.ylabel('dB')
# plt.show()
# plt.close('all')



# # TEST TRANSIT CYGA
# b = BST('/Users/aloh/Desktop/NenuFAR_Data/20181018_195600_BST.fits')
# b.select(freq=60)
# t=Transit()
# t.from_bst(obs='/Users/aloh/Desktop/NenuFAR_Data/20181018_195600_BST.fits', dt=60, resol=0.2, freq=60)
# x = (t.simulation['time'] - t.simulation['time'][0]).sec / 60
# y = t.simulation['amp']
# scale = np.median(b.data['amp']) / np.median(y)
# plt.plot(x, 10 * np.log10(y))
# plt.plot( (b.data['time']-t.simulation['time'][0]).sec/60, 10 * np.log10(b.data['amp']/scale) )
# plt.xlabel('Time (min since {})'.format(t.simulation['time'][0].iso))
# plt.ylabel('dB')
# plt.show()
# plt.close('all')


# TEST TRANSIT CYGA 2 MINIARRAYS
t = Transit()
t.from_bst(obs='/Users/aloh/Desktop/NenuFAR_Data/20190228_071900_BST.fits', dt=600, freq=60, resol=0.2)
b = BST('/Users/aloh/Desktop/NenuFAR_Data/20190228_071900_BST.fits')
b.select(freq=60)
x = (t.simulation['time'] - t.simulation['time'][0]).sec / 60
y = t.simulation['amp']
scale = np.median(b.data['amp']) / np.median(y)
plt.plot(x, 10 * np.log10(y*scale))
plt.plot( (b.data['time']-t.simulation['time'][0]).sec/60, 10 * np.log10(b.data['amp']) )
plt.ylim((69.5, 74))
plt.xlabel('Time (min since {})'.format(t.simulation['time'][0].iso))
plt.ylabel('dB')
plt.show()
plt.close('all')


# # TEST TRANSIT SST ANT UNIQUE ZENITH
# from astropy.io import fits
# from astropy.time import Time
# FREQ=60
# files = ['/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_140000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_150000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_160000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_170000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_180000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_190000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_200000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_210000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_220000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161020_230000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_000000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_010000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_020000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_030000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_040000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_050000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_060000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_070000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_080000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_090000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_100000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_110000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_120000_MR0_NW_SST.fits',
#     '/Users/aloh/Documents/Work/NenuFAR/Tests/20161021_130000_MR0_NW_SST.fits']
# freqs = np.arange(0.09765625, 99.90234375+0.1953125, 0.1953125)
# f_ind = (np.abs(freqs - FREQ)).argmin()
# for ff in sorted(files):
#     data = fits.getdata(ff, ext=1, memmap=True)
#     if 'mr0data' in locals():
#         mr0data = np.hstack( (mr0data, data['data'][:, f_ind]) )
#         mr0time = np.hstack( (mr0time, data['jd']) )
#     else :
#         mr0data = data['data'][:, f_ind]
#         mr0time = data['jd']
# mr0time = Time(mr0time, format='jd')
# t=Transit()
# t.predict(src=None, resol=1, time=mr0time[0], freq=FREQ, duration=86400, miniarrays=[1], azana=180, elana=90, azdig=180, eldig=90, dt=3600, polar='NW')
# x = (t.simulation['time'] - t.simulation['time'][0]).sec / 3600
# y = t.simulation['amp']
# scale = np.median(mr0data) / np.median(y)
# plt.plot( (mr0time-t.simulation['time'][0]).sec/3600, 10 * np.log10(mr0data) )
# plt.plot(x, 10 * np.log10(y*scale))
# plt.xlabel('Time (min since {})'.format(t.simulation['time'][0].iso))
# plt.ylabel('dB')
# plt.show()
# plt.close('all')


# # TEST TRANSIT SST ZENITH
# import glob
# FREQ=60
# for mr in range(40):
#     MR=mr
#     sst = SST(sorted(glob.glob('/Users/aloh/Documents/Work/NenuFAR/Nenupy3_Tests/Zenith_SST/*SST.fits'))[0:23])
#     sst.select(freq=FREQ, ma=mr)
#     t=Transit()
#     t.predict(src=None, resol=0.4, time=sst.data['time'][0], freq=FREQ, duration=86400, miniarrays=[MR], azana=180, elana=90, azdig=180, eldig=90, dt=1800, polar='NW')
#     x = (t.simulation['time'] - t.simulation['time'][0]).sec / 3600
#     y = t.simulation['amp']
#     scale = np.median(sst.data['amp']) / np.median(y)
#     plt.plot( (sst.data['time']-t.simulation['time'][0]).sec/3600, 10 * np.log10(sst.data['amp']) )
#     plt.plot(x, 10 * np.log10(y*scale))
#     plt.xlabel('Time (min since {})'.format(t.simulation['time'][0].iso))
#     plt.ylabel('dB')
#     plt.savefig('/Users/aloh/Documents/Work/NenuFAR/Nenupy3_Tests/Zenith_SST/mr{}_zenith_60mhz.png'.format(MR))
#     # plt.show()
#     plt.close('all')





#! /usr/bin/python3.6
# -*- coding: utf-8 -*-

"""
Class to handle pygsm sky models
        by A. Loh
"""

import os
import sys
import numpy as np

import healpy as hp
from matplotlib import pyplot as plt
import scipy.ndimage as ndimage

try:
    from pygsm import GlobalSkyModel
except:
    print("\n\t=== WARNING: PyGSM is not installed ===")
    # raise ImportError("\n\t=== Impossible to import pygsm ===")    

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['SkyModel']

class SkyModel(object):
    def __init__(self):
        self.nra  = 1024
        self.ndec = 512
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def pointSource(self, ra, dec, intensity=1.e6, offset=0.1, sigma=1):
        """ Generate a fake sky model in equatorial coordinates
            Empty sky with a single bright pixel

            Parameters
            ----------
            ra : float
                Right-Ascension of the point source in degrees 
            dec : float
                Declination of the point source in degrees
            intensity : float, optional
                Value of the point-source (default=1.e6)
            offset : float, optional
                Value of sky background (default=0.1)
            sigma : float, optional
                sigma value in pixels for Gaussian smoothing of the point source (default=0)

            Returns
            -------
            self.skymodel : np.ndarray
                Sky model
        """
        self.skymodel = np.zeros( (self.ndec, self.nra) ) + offset
        decs = np.linspace(-90., 90.,   self.ndec) 
        ras  = np.linspace(180., -180., self.nra)
        if ra > 180.:
            ra -= 360. 
        decind  = min(range(decs.size), key=lambda i: np.abs(decs[i] - dec))
        raind   = min(range(ras.size),  key=lambda i: np.abs(ras[i]  - ra))
        self.skymodel[decind, raind] = intensity
        if sigma != 0.:
            self.skymodel = ndimage.gaussian_filter(self.skymodel, sigma=(sigma, sigma), order=0, mode='wrap')
        return

    def gsm2008(self, freq, healpix=False):
        """ Generate a PyGSM 2008 skymodel in equatorial coordinates

            Parameters
            ----------
            freq : float
                Frequency in MHz

            Returns
            -------
            self.skymodel : np.ndarray
                Sky model
        """
        gsm     = GlobalSkyModel(freq_unit='MHz')
        gsmmap  = gsm.generate(freq)
        if not healpix:
            gsmcart = hp.cartview(gsmmap, coord=['G', 'C'], xsize=self.nra,
                ysize=self.ndec, return_projected_map=True)
            plt.close('all')
        else:
            gsmcart = gsmmap
        self.skymodel = gsmcart
        return











#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
"""

import os
import numpy as np
import healpy as hp

from nenupy.beam.antenna import miniarrays


__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2019, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['Anabeam', 'Digibeam']


# ============================================================= #
# -------------------------- Anabeam -------------------------- #
# ============================================================= #
class Anabeam(object):
    """
    """
    
    def __init__(self, **kwargs):
        self.instrument_effect = True
        self.antpos_offset = 90 # angle to rotate
        self._anakwargs(kwargs)

    # ========================================================= #
    # --------------------- Getter/Setter --------------------- #
    @property
    def freq(self):
        """ Frequency selection in MHz
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        assert 0. <= f <= 100.,\
            '{} MHz outside the frequency range'.format(f)
        self._freq = f
        self.kwave = 2 * np.pi * f * 1.e6 / 299792458
        return
    # --------------------------------------------------------- #
    @property
    def polar(self):
        """ Polarization selection
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        assert p.upper() in ['NW', 'NE'],\
            'Polar {} not understood (NW or NE)'
        self._polar = p.upper()
        return
    # --------------------------------------------------------- #
    @property
    def azana(self):
        """ Pointed azimuth in degrees
        """
        return self._azana
    @azana.setter
    def azana(self, a):
        self._azana = np.radians(a)
        return
    # --------------------------------------------------------- #
    @property
    def elana(self):
        """ Pointed elevation in degrees
        """
        return self._elana
    @elana.setter
    def elana(self, e):
        e = self._squint(e)
        self._elana = np.radians(e)
        return
    # --------------------------------------------------------- #
    @property
    def ma(self):
        """ Miniarray selection
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        marecorded = miniarrays.ma[:, 0].astype(int)
        assert m in marecorded,\
            'MA {} not recorded'.format(m)
        self._ma = m
        return
    # --------------------------------------------------------- #
    @property
    def marot(self):
        """ MA rotation selection
            The '0' seems to be shifted by 90 deg (conventions...)
        """
        return self._marot[self.ma]
    @marot.setter
    def marot(self, r):
        if r is None:
            self._marot = miniarrays.ma[:, 1] + self.antpos_offset 
        else:
            self._marot = np.repeat(r, miniarrays.ma[:, 1].size) + self.antpos_offset
        return
    # --------------------------------------------------------- #
    @property
    def resol(self):
        """ Angular resolution in degrees
        """
        return self._resol
    @resol.setter
    def resol(self, r):
        if r < 0.1:
            print('WARNING, high-res (<0.1deg) may result in long computational durations')
        # Find the nside corresponding to the desired resolution
        nsides = np.array([2**i for i in range(1, 12)])
        resol_rad = hp.nside2resol(nside=nsides, arcmin=False)
        resol_deg = np.degrees(resol_rad)
        idx = (np.abs(resol_deg - r)).argmin()
        if hasattr(self, 'resol'):
            if resol_deg[idx] == self.resol:
                # No need to recompute it
                return
        self._resol = resol_deg[idx]
        self._nside = nsides[idx]
        self._npix = hp.nside2npix(self._nside)
        self._skygrid()
        return


    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def get_anabeam(self):
        """
        """
        beam = self._compute_anabeam()
        return


    # ========================================================= #
    # ----------------------- Internal ------------------------ #
    def _skygrid(self):
        """ Compute the sky grid
        """
        self.azgrid, self.elgrid = hp.pix2ang(nside=self._nside,
                                    ipix=np.arange(self._npix),
                                    lonlat=True,
                                    nest=False)
        self.azgrid = np.radians(self.azgrid[::-1]) # east is 90
        self.elgrid = np.radians(self.elgrid)
        self._over_horizon = (self.elgrid < 0.)
        self.domega = hp.nside2resol(self._nside)**2. * np.cos(np.pi/2 - self.elgrid)
        self.domega[self._over_horizon] = 0.
        return
    # --------------------------------------------------------- #
    def _ant_pos(self):
        """ Return the antenna position within a mini-array
        """
        antpos = miniarrays.antpos
        if self.marot is not None:
            rot = np.radians( self.marot )
            rotation = np.array([[ np.cos(rot), np.sin(rot), 0],
                                 [-np.sin(rot), np.cos(rot), 0],
                                 [ 0,           0,           1]])
            antpos = np.dot( antpos, rotation )
        return antpos#[9].reshape((1, 3)) # single_ant
    # --------------------------------------------------------- #
    def _ant_gain(self):
        """
        """
        if hasattr(self, 'antgain'):
            if hp.npix2nside(self.antgain.size) == self._nside:
                # This has already been computed
                return self.antgain
        f1 = int( np.floor( self.freq/10. ) ) * 10
        f2 = int( np.ceil(  self.freq/10. ) ) * 10
        cols={'NW_10': 0,
              'NW_20': 1,
              'NW_30': 2,
              'NW_40': 3,
              'NW_50': 4,
              'NW_60': 5,
              'NW_70': 6,
              'NW_80': 7,
              'NE_10': 8,
              'NE_20': 9,
              'NE_30': 10,
              'NE_40': 11,
              'NE_50': 12,
              'NE_60': 13,
              'NE_70': 14,
              'NE_80': 15}
        modulpath = os.path.dirname(os.path.realpath(__file__))
        antfile = os.path.join(modulpath, 'NenuFAR_Ant_Hpx.fits') 
        antgain1 = hp.read_map(filename=antfile,
            hdu=1,
            field=cols['{}_{}'.format(self.polar, f1)],
            verbose=False,
            memmap=True)
        antgain2 = hp.read_map(filename=antfile,
            hdu=1,
            field=cols['{}_{}'.format(self.polar, f2)],
            verbose=False,
            memmap=True)
        if f1 != f2:
            antgain = antgain1 * (f2-self.freq) / 10. + antgain2 * (self.freq-f1) / 10.
        else:
            antgain = antgain1
        self.antgain = hp.ud_grade(antgain, nside_out=self._nside)
        return self.antgain
    # --------------------------------------------------------- #
    def _real_pointing(self, azimuth, elevation):
        """ azimuth and elevation in radians
        """
        if self.instrument_effect:
            from astropy.io.fits import getdata
            ff = '/Users/aloh/Documents/Work/Scripts/Own_Packages/nenupy/nenupy/beam/NenuFAR_thph.fits'
            thph = getdata( ff )
            phi = int(np.degrees(azimuth)/0.05 - 0.5)  
            theta = int((90.-np.degrees(elevation))/0.05 - 0.5)
            t, p = thph[:, theta, phi]
            azimuth = np.radians(p)
            elevation = np.radians(90. - t)
        return azimuth, elevation
    # --------------------------------------------------------- #
    def _squint(self, elevation):
        """ elevation in degrees
        """
        if self.instrument_effect:
            from scipy.io.idl import readsav
            from scipy.interpolate import interp1d
            squint  = readsav('/Users/aloh/Documents/Work/Scripts/Own_Packages/nenupy/nenupy/beam/squint_table.sav')
            optfreq = 30
            indfreq = np.where(squint['freq']==optfreq)[0][0]
            newele  = interp1d(squint['elev_desiree'][indfreq,:], squint['elev_a_pointer'])(elevation)
            if newele < 20.: # squint is limited at 20 deg elevation
                newele = 20.
            elevation = newele
        return elevation    
    # --------------------------------------------------------- #
    def _compute_anabeam(self):
        """ Compute e^(i (phi0 - phi))
            Where: phi0 is the MA reference phase
                   phi is the phase towards the pointing
        """
        antpos = self._ant_pos()

        az, el = self._real_pointing(self.azana, self.elana)
        ux = np.cos(az) * np.cos(el)
        uy = np.sin(az) * np.cos(el)
        uz = np.sin(el)
        phix = antpos[:, 0] * ux[np.newaxis]
        phiy = antpos[:, 1] * uy[np.newaxis]
        phiz = antpos[:, 2] * uz[np.newaxis]
        phi  = phix + phiy + phiz

        # ------ Phase reference ------ #
        ux = np.cos(self.azgrid) * np.cos(self.elgrid) 
        uy = np.sin(self.azgrid) * np.cos(self.elgrid) 
        uz = np.sin(self.elgrid)
        phi0x = antpos[:, 0] * ux[:, np.newaxis]
        phi0y = antpos[:, 1] * uy[:, np.newaxis]
        phi0z = antpos[:, 2] * uz[:, np.newaxis]
        phi0  = phi0x + phi0y + phi0z
        
        dphase = self.kwave * (phi0[:, :] - phi[np.newaxis, :])

        # ------ e^(i Phi) ------ #
        eiphi = np.sum( np.exp(1j * dphase), axis=1 )

        beam = eiphi * eiphi.conjugate() * self._ant_gain()
        # beam = self._ant_gain()
        beam = np.real(beam)
        beam[self._over_horizon] = 0
        self.anabeam = beam# / beam.max()
        return
    # --------------------------------------------------------- #
    def _anakwargs(self, kwargs):
        """ Evaluate the keyword arguments and fill attributes
        """
        self.ma    = kwargs.get('ma', 0)
        self.marot = kwargs.get('marot', None)
        self.freq  = kwargs.get('freq', 50)
        self.polar = kwargs.get('polar', 'NW')
        self.azana = kwargs.get('azana', 180.)
        self.elana = kwargs.get('elana', 90.)
        self.resol = kwargs.get('resol', 0.9)
        return
# ============================================================= #
# ============================================================= #







# ============================================================= #
# ------------------------- Digibeam -------------------------- #
# ============================================================= #
class Digibeam(Anabeam):
    """
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._digikwargs(kwargs)

    # ========================================================= #
    # --------------------- Getter/Setter --------------------- #
    @property
    def miniarrays(self):
        return self._miniarrays
    @miniarrays.setter
    def miniarrays(self, m):
        marecorded = miniarrays.ma[:, 0].astype(int)
        if m is None:
            m = marecorded
        if not isinstance(m, (list, np.ndarray)):
            m = [m]
        assert all([mi in marecorded for mi in m]),\
            'Invalid MA index/indices.'
        self._miniarrays = np.array(m)
        return
    # --------------------------------------------------------- #
    @property
    def mapos(self):
        """ MA position selection [[x0,y0,z0], [x1,y1,...]]
        """
        return self._mapos[self.miniarrays]
    @mapos.setter
    def mapos(self, p):
        if p is None:
            self._mapos = miniarrays.ma[:, 2:5]
        else:
            self._mapos = np.array(p)
            assert self._mapos.shape[1] == 3,\
                'Positions should have 3 coordinates.'
            assert self._mapos.shape[0] == self.miniarrays.size,\
                'No of MAs not equal to No of positions.' 
        return
    # --------------------------------------------------------- #
    @property
    def azdig(self):
        """ Pointed azimuth in degrees
        """
        return self._azdig
    @azdig.setter
    def azdig(self, a):
        self._azdig = np.radians(a)
        return
    # --------------------------------------------------------- #
    @property
    def eldig(self):
        """ Pointed elevation in degrees
        """
        return self._eldig
    @eldig.setter
    def eldig(self, e):
        self._eldig = np.radians(e)
        return

    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def get_digibeam(self):
        """
        """
        # import pylab as plt

        self._compute_digibeam()

        # hp.cartview( np.log10(self.digibeam))
        # # hp.graticule()
        # plt.show()

            
        return self.digibeam

    # ========================================================= #
    # ----------------------- Internal ------------------------ #
    def _compute_digibeam(self):
        """
        """
        # Build the Mini-Array 'summed' response
        abeams = {}
        for ma in self.miniarrays:
            self.ma = ma
            if str(self.marot%60) not in abeams.keys():
                self._compute_anabeam()
                abeams[str(self.marot%60)] = self.anabeam.copy()
                if self.miniarrays.size == 1:
                    # Only take the Anabeam
                    self.digibeam = self.anabeam
                    return
            if not 'summed_mas' in locals():
                summed_mas = abeams[str(self.marot%60)]
            else:
                summed_mas += abeams[str(self.marot%60)]

        ux = np.cos(self.azdig) * np.cos(self.eldig)
        uy = np.sin(self.azdig) * np.cos(self.eldig)
        uz = np.cos(self.eldig)
        phix = self.mapos[:, 0] * ux[np.newaxis]
        phiy = self.mapos[:, 1] * uy[np.newaxis]
        phiz = self.mapos[:, 2] * uz[np.newaxis]
        phi  = phix + phiy + phiz

        # ------ Phase reference ------ #
        ux = np.cos(self.azgrid) * np.cos(self.elgrid) 
        uy = np.sin(self.azgrid) * np.cos(self.elgrid) 
        uz = np.sin(self.elgrid)
        phi0x = self.mapos[:, 0] * ux[:, np.newaxis]
        phi0y = self.mapos[:, 1] * uy[:, np.newaxis]
        phi0z = self.mapos[:, 2] * uz[:, np.newaxis]
        phi0  = phi0x + phi0y + phi0z
        
        dphase = self.kwave * (phi0[:, :] - phi[np.newaxis, :])

        # ------ e^(i Phi) ------ #
        eiphi = np.sum( np.exp(1j * dphase), axis=1 )

        beam = eiphi * eiphi.conjugate() * summed_mas
        beam = np.real(beam)
        # beam[self._over_horizon] = 0.
        self.digibeam = beam# / beam.max()

        return
    # --------------------------------------------------------- #
    def _digikwargs(self, kwargs):
        """ Evaluate the keyword arguments and fill attributes
        """
        self.miniarrays = kwargs.get('miniarrays', None)
        self.mapos      = kwargs.get('mapos', None)
        self.azdig      = kwargs.get('azdig', 180.)
        self.eldig      = kwargs.get('eldig', 90.)
        return


# ============================================================= #
# ============================================================= #






from nenupy.hpx import Digibeam, Skymodel
import healpy as hp
import matplotlib as mpl
import pylab as plt
import matplotlib.ticker as mtick
import numpy as np


db = Digibeam(azana=180,
            azdig=180,
            elana=75,
            eldig=75,
            miniarrays=[10, 20, 30, 40],
            freq=60,
            polar='NW',
            resol=0.2)

beam = db.get_digibeam()

print(db._nside)

sm = Skymodel(nside=db._nside,
            freq=60)

skybeam = beam * sm.get_skymodel(time='2019-03-15 12:00:00', model='GSM')


hp.mollview(skybeam, flip='astro', title=None, max=skybeam.max(), min=skybeam.max()*1e-5, norm='log')
hp.graticule()
plt.show()




#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
"""

import os, sys
import numpy as np

import healpy as hp
from astropy.time import Time

from pygsm import GlobalSkyModel

from nenupy.astro import toRadec, getSrc


__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2019, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['Skymodel']


# ============================================================= #
# ------------------------- Skymodel -------------------------- #
# ============================================================= #
class Skymodel(object):
    """
    """

    def __init__(self, **kwargs):
        self._kwargs(kwargs)

    # ========================================================= #
    # --------------------- Getter/Setter --------------------- #
    @property
    def time(self):
        return self._time
    @time.setter
    def time(self, t):
        if not isinstance(t, Time):
            if t.lower() == 'now':
                t = Time.now()
            else:
                t = Time(t)
        self._time = t
        return


    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def get_skymodel(self, time='now', model='gsm', **kwargs):
        """ Get the GMS sky at a given time and frequency
            above NenuFAR
        """
        self.time = time

        if model.lower() == 'gsm': 
            if not hasattr(self, 'map'):
                self._load_gsm_model()
        elif model.lower() == 'pointsource':
            if not hasattr(self, 'map'):
                try:
                    ra = kwargs['ra']
                    dec = kwargs['dec']
                except:
                    raise ValueErrror('ra and dec should be specified.')
                self._load_point_model(ra=ra, dec=dec)
        else:
            raise ValueErrror('Only gsm or pointsource')

        radec = toRadec((0., 0.),
            time=self.time,
            loc='NenuFAR',
            unit='rad')

        rot_dec = radec.dec.rad
        rot_ra = (radec.ra.rad)%(2.*np.pi)

        rot = hp.Rotator(deg=False,
            rot=[rot_ra, rot_dec],
            coord=['G', 'C'])

        sys.stdout = open(os.devnull, 'w')
        skymap = rot.rotate_map_alms(self.map)
        # skymap = rot.rotate_map_pixel(self.map)
        sys.stdout = sys.__stdout__

        return skymap

    # ========================================================= #
    # ----------------------- Internal ------------------------ #
    def _load_gsm_model(self):
        """
        """
        gsm   = GlobalSkyModel(freq_unit='MHz')
        gsmap = gsm.generate(self.freq)
        self.map = hp.pixelfunc.ud_grade(gsmap,
            nside_out=self.nside)
        return
    # --------------------------------------------------------- #
    def _load_point_model(self, ra, dec):
        """ ra, dec in radians
        """
        src = getSrc(source=(ra, dec))
        src_gal = src.galactic
        self.map = np.repeat(1.e-1, hp.nside2npix(self.nside))
        pixel = hp.ang2pix(nside=self.nside,
            theta=src_gal.l.deg,
            phi=src_gal.b.deg,
            lonlat=True,
            nest=False)
        self.map[pixel] = 1.e6
        return
    # --------------------------------------------------------- #
    def _kwargs(self, kwargs):
        """ Evaluate the keyword arguments and fill attributes
        """
        self.nside = kwargs.get('nside', 256)
        self.freq = kwargs.get('freq', 50)
        return
# ============================================================= #
# ============================================================= #



#! /usr/bin/python3.6
# -*- coding: utf-8 -*-

"""
Class Simulate SST/BST time profiles
        by A. Loh


from nenupy.read import SST
from nenupy.beam import SSTbeam
from nenupy.skymodel import SkyModel
from nenupy.simulation import Transit
sst = SST('20170425_230000_SST.fits')
sst.select(freq=50)
beam = SSTbeam(sst)
sm = SkyModel()
sm.gsm2008(freq=sst.freq)
simul = Transit(obs=sst, beam=beam, skymodel=sm)
simul.dt = 300
simul.plotProfile()

from nenupy.read import BST
from nenupy.beam import BSTbeam
from nenupy.skymodel import SkyModel
from nenupy.simulation import Transit
bst = BST('20181018_195600_BST.fits')
bst.select(freq=50)
sm = SkyModel()
sm.gsm2008(freq=bst.freq)
beam = BSTbeam(bst)
simul = Transit(obs=bst, beam=beam, skymodel=sm)
simul.dt = 60
simul.plotProfile()

"""

import os
import sys
import numpy as np
# import matplotlib as mpl
from matplotlib import pyplot as plt
from scipy.interpolate import interp2d

from astropy.time import Time, TimeDelta
from astropy import units as u
from astropy import coordinates as coord

from ..read import SST, BST
from ..beam import SSTbeam, BSTbeam
from ..skymodel import SkyModel
from ..beam.antenna import miniarrays
from ..utils import ProgressBar

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['Transit', 'Tracking']


class Transit():
    def __init__(self, obs, beam, skymodel, start=None, stop=None):
        self.obs      = obs
        self.beam     = beam
        self.skymodel = skymodel
        self.dt       = TimeDelta(60, format='sec')
        self.start    = start
        self.stop     = stop

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def obs(self):
        """ Observation (SST or BST)
        """
        return self._obs
    @obs.setter
    def obs(self, o):
        if not isinstance(o, (SST, BST)):
            raise AttributeError("\n\t=== obs must be either a SST or a BST class ===")
        else:
            assert o.type == 'transit', 'Observation doesnt look like a transit'

            if not hasattr(o, 'data'):
                # Select a default time-profile
                print("\n\t=== WARNING: default time profile ===")
                o.select(time=[o.obstart, o.obstop], freq=50, polar='nw')
            self._obs = o
        return

    @property
    def beam(self):
        """ Simulated beam
        """
        return self._beam
    @beam.setter
    def beam(self, b):
        if not isinstance(b, (SSTbeam, BSTbeam)):
            raise AttributeError("\n\t=== beam must be either a SSTbeam or a BSTbeam class ===")
        else:
            if not hasattr(b, 'beam'):
                # Compute the beam
                b.getBeam()
            self._beam = b
        return

    @property
    def skymodel(self):
        """ Sky model
        """
        return self._skymodel
    @skymodel.setter
    def skymodel(self, s):
        if not isinstance(s, SkyModel):
            raise AttributeError("\n\t=== skymodel must be a SkyModel class ===")
        else:
            self._skymodel = s
        return

    @property
    def dt(self):
        """ Simulation time resolution in seconds
        """
        return self._dt
    @dt.setter
    def dt(self, d):
        if not isinstance(d, TimeDelta):
            try:
                d = TimeDelta(d, format='sec')
            except:
                raise AttributeError("\n\t=== 'dt' format not understood ===")
        self._dt = d
        return

    @property
    def start(self):
        """ Simulation time start
        """
        return self._start
    @start.setter
    def start(self, t):
        if t is None:
            t = self.obs.data['time'][0]

        if not isinstance(t, Time):
            try:
                t = Time(t)
            except:
                raise AttributeError("\n\t=== 'start' format not understood ===")
        self._start = t
        return

    @property
    def stop(self):
        """ Simulation time stop
        """
        return self._stop
    @stop.setter
    def stop(self, t):
        if t is None:
            t = self.obs.data['time'][-1]

        if not isinstance(t, Time):
            try:
                t = Time(t)
            except:
                raise AttributeError("\n\t=== 'stop' format not understood ===")
        self._stop = t
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getProfile(self):
        """ Compute the time-profile simulation
        """
        # ------ Find number of steps ------ #
        nbtime  = int(np.ceil( (self.stop - self.start).sec / self.dt.sec ))
        self.dt = TimeDelta( (self.stop - self.start).sec / (nbtime-1), format='sec')
        
        self.f = self.obs.freq
        self.t = np.zeros( nbtime )
        self.d = np.zeros( nbtime ) 

        # ------ Load skymodel ------ #
        skymodel = self.skymodel.skymodel
        ramodel  = np.linspace(180., -180., skymodel.shape[1]) 
        decmodel = np.linspace(-90., 90., skymodel.shape[0]) 
        azimuth  = np.linspace(0., 360., skymodel.shape[1])
        zenitha  = 90. - np.linspace(0., 90., skymodel.shape[0]/2)
        agrid, zgrid = np.meshgrid( azimuth, zenitha )
        domega = np.radians(ramodel[0]-ramodel[1])**2. * np.cos(np.radians(zgrid))
        #domega /= domega.max()

        # ------ Fixed beam ------ #
        beam = np.flipud(self.beam.beam)
        bazi = np.linspace(0., 360., beam.shape[1])
        bele = np.linspace(0., 90.,  beam.shape[0])
        fb   = interp2d(bazi, bele, beam, kind='linear' )
        beam = fb( np.linspace(0, 360, skymodel.shape[1]), np.linspace(0, 90, skymodel.shape[0]/2) )

        # ------ Loop over time ------ #
        bar = ProgressBar(valmax=nbtime, title='Transit observation simulation')
        for i in range(nbtime):
            # ------ Rotate skymodel ------ #
            frame  = coord.AltAz(obstime=self.start + i*self.dt, location=miniarrays.nenufarloc)

            altaz  = coord.SkyCoord(agrid*u.deg, (90-zgrid)*u.deg, frame=frame)
            radec  = altaz.transform_to(coord.FK5(equinox='J2000'))
            ragrid, decgrid = radec.ra.deg, radec.dec.deg
            ragrid[ragrid > 180] -= 360 # RA is within -180 and 180
            xraind  = (ramodel[0] - ragrid)   / (ramodel[0]-ramodel[1])   #+ 0.5
            xdecind = (decgrid - decmodel[0]) / (decmodel[1]-decmodel[0]) #+ 0.5
            skysel  = skymodel[xdecind.astype(int), xraind.astype(int)]

            # ------ Integrate ------ #
            amp       = np.sum(skysel * beam * domega) # / (4 * np.pi) # domega is normalised
            self.d[i] = amp
            self.t[i] = (self.start + i*self.dt).mjd
            bar.update()
        return

    def plotProfile(self):
        """ Plot the simulation profile against the data
        """
        if not hasattr(self, 'd'):
            self.getProfile()
        tmask = (self.obs.data['time'] >= self.start) & (self.obs.data['time'] <= self.stop)
        scale = np.median(self.obs.data['amp'][tmask]) / np.median(self.d)
        plt.plot( (self.obs.data['time'][tmask] - self.start).sec/60., self.obs.data['amp'][tmask], label='Observation')
        plt.plot( (Time(self.t, format='mjd') - self.start).sec/60., self.d * scale, label='Simulation')
        plt.xlabel('Time (min since {})'.format(self.start.iso))
        plt.ylabel('Amplitude')
        plt.show()
        plt.close('all')
        return

    def saveProfile(self, savefile=None):
        """ Save the simulation
        """
        if savefile is None:
            savefile = self.obs.obsname + '_simulation.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        if not hasattr(self, 'd'):
            self.getProfile()

        prihdr = fits.Header()
        prihdr.set('OBS', self.obs.obsname)
        prihdr.set('FREQ', str(self.obs.freq))
        prihdr.set('TIME', str(self.obs.time))
        prihdr.set('POLAR', self.polar)
        # prihdr.set('MINI-ARR', str(self.ma))
        datahdu = fits.PrimaryHDU(self.d.T, header=prihdr)
        freqhdu = fits.BinTableHDU.from_columns( [fits.Column(name='frequency', format='D', array=self.f)] )
        timehdu = fits.BinTableHDU.from_columns( [fits.Column(name='mjd', format='D', array=self.t)] )
        hdulist = fits.HDUList([datahdu, freqhdu, timehdu])
        hdulist.writeto(savefile, overwrite=True)
        return



class Tracking():
    def __init__(self, obs, skymodel, start=None, stop=None):
        self.obs      = obs
        self.skymodel = skymodel
        self.dt       = TimeDelta(60, format='sec')
        self.start    = start
        self.stop     = stop

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def obs(self):
        """ Observation (SST or BST)
        """
        return self._obs
    @obs.setter
    def obs(self, o):
        if not isinstance(o, (SST, BST)):
            raise AttributeError("\n\t=== obs must be either a SST or a BST class ===")
        else:
            assert o.type == 'tracking', 'Observation doesnt look like a tracking'

            if not hasattr(o, 'd'):
                # Select a default time-profile
                print("\n\t=== WARNING: default time profile ===")
                o.select(time=[o.obstart, o.obstop], freq=50, polar='nw')
            self._obs = o
        return

    @property
    def skymodel(self):
        """ Sky model
        """
        return self._skymodel
    @skymodel.setter
    def skymodel(self, s):
        if not isinstance(s, SkyModel):
            raise AttributeError("\n\t=== skymodel must be a SkyModel class ===")
        else:
            self._skymodel = s
        return

    @property
    def dt(self):
        """ Simulation time resolution in seconds
        """
        return self._dt
    @dt.setter
    def dt(self, d):
        if not isinstance(d, TimeDelta):
            try:
                d = TimeDelta(d, format='sec')
            except:
                raise AttributeError("\n\t=== 'dt' format not understood ===")
        self._dt = d
        return

    @property
    def start(self):
        """ Simulation time start
        """
        return self._start
    @start.setter
    def start(self, t):
        if t is None:
            t = self.obs.data['time'][0]

        if not isinstance(t, Time):
            try:
                t = Time(t)
            except:
                raise AttributeError("\n\t=== 'start' format not understood ===")
        self._start = t
        return

    @property
    def stop(self):
        """ Simulation time stop
        """
        return self._stop
    @stop.setter
    def stop(self, t):
        if t is None:
            t = self.obs.data['time'][-1]

        if not isinstance(t, Time):
            try:
                t = Time(t)
            except:
                raise AttributeError("\n\t=== 'stop' format not understood ===")
        self._stop = t
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getProfile(self):
        """ Compute the time-profile simulation
        """
        # ------ Find number of steps ------ #
        nbtime  = int(np.ceil( (self.stop - self.start).sec / self.dt.sec ))
        self.dt = TimeDelta( (self.stop - self.start).sec / (nbtime-1), format='sec')
        
        self.f = self.obs.freq
        self.t = np.zeros( nbtime )
        self.d = np.zeros( nbtime )
        tmp_coords = (None, None, None, None) 

        # ------ Load skymodel ------ #
        skymodel = self.skymodel.skymodel
        ramodel  = np.linspace(180., -180., skymodel.shape[1]) 
        decmodel = np.linspace(-90., 90., skymodel.shape[0]) 
        azimuth  = np.linspace(0., 360., skymodel.shape[1])
        zenitha  = 90. - np.linspace(0., 90., skymodel.shape[0]/2)
        agrid, zgrid = np.meshgrid( azimuth, zenitha )
        domega = np.radians(ramodel[0]-ramodel[1])**2. * np.cos(np.radians(zgrid))
        #domega /= domega.max()

        # ------ Loop over time ------ #
        bar = ProgressBar(valmax=nbtime, title='Tracking observation simulation')
        for i in range(nbtime):
            # ------ Re-compute the beam ------ #
            self.obs.time = self.start + i*self.dt
            if ((self.obs.azana != tmp_coords[0]) or
                (self.obs.elana != tmp_coords[1]) or
                (self.obs.azdig != tmp_coords[2]) or
                (self.obs.eldig != tmp_coords[3])):
                if isinstance(self.obs, SST):
                    pass
                elif isinstance(self.obs, BST):
                    obsbeam = BSTbeam(self.obs)
                    obsbeam.getBeam()
                beam = np.flipud(obsbeam.beam)
                bazi = np.linspace(0., 360., beam.shape[1])
                bele = np.linspace(0., 90.,  beam.shape[0])
                fb   = interp2d(bazi, bele, beam, kind='linear' )
                beam = fb( np.linspace(0, 360, skymodel.shape[1]), np.linspace(0, 90, skymodel.shape[0]/2) )
                tmp_coords = (self.obs.azana.copy(),
                              self.obs.elana.copy(),
                              self.obs.azdig.copy(),
                              self.obs.eldig.copy())

            # ------ Rotate skymodel ------ #
            frame  = coord.AltAz(obstime=self.start + i*self.dt, location=miniarrays.nenufarloc)

            altaz  = coord.SkyCoord(agrid*u.deg, (90-zgrid)*u.deg, frame=frame)
            radec  = altaz.transform_to(coord.FK5(equinox='J2000'))
            ragrid, decgrid = radec.ra.deg, radec.dec.deg
            ragrid[ragrid > 180] -= 360 # RA is within -180 and 180
            xraind  = (ramodel[0] - ragrid)   / (ramodel[0]-ramodel[1])   #+ 0.5
            xdecind = (decgrid - decmodel[0]) / (decmodel[1]-decmodel[0]) #+ 0.5
            skysel  = skymodel[xdecind.astype(int), xraind.astype(int)]

            # ------ Integrate ------ #
            amp       = np.sum(skysel * beam * domega) # / (4 * np.pi) # domega is normalised
            self.d[i] = amp
            self.t[i] = (self.start + i*self.dt).mjd
            bar.update()
        return

    def plotProfile(self, showdata=True):
        """ Plot the simulation profile against the data
        """
        if not hasattr(self, 'd'):
            self.getProfile()
        
        if showdata:
            tmask = (self.obs.data['time'] >= self.start) & (self.obs.data['time'] <= self.stop)
            scale = np.median(self.obs.data['amp'][tmask]) / np.median(self.d)
            plt.plot( (self.obs.data['time'][tmask] - self.start).sec/60., self.obs.data['amp'][tmask], label='Observation')
        else:
            scale = 1.
        plt.plot( (Time(self.t, format='mjd') - self.start).sec/60., self.d * scale, label='Simulation')
        plt.xlabel('Time (min since {})'.format(self.start.iso))
        plt.ylabel('Amplitude')
        plt.show()
        plt.close('all')
        return

    def saveProfile(self, savefile=None):
        """ Save the simulation
        """
        if savefile is None:
            savefile = self.obs.obsname + '_simulation.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        if not hasattr(self, 'd'):
            self.getProfile()

        prihdr = fits.Header()
        prihdr.set('OBS', self.obs.obsname)
        prihdr.set('FREQ', str(self.obs.freq))
        prihdr.set('TIME', str(self.obs.time))
        prihdr.set('POLAR', self.polar)
        # prihdr.set('MINI-ARR', str(self.ma))
        datahdu = fits.PrimaryHDU(self.d.T, header=prihdr)
        freqhdu = fits.BinTableHDU.from_columns( [fits.Column(name='frequency', format='D', array=self.f)] )
        timehdu = fits.BinTableHDU.from_columns( [fits.Column(name='mjd', format='D', array=self.t)] )
        hdulist = fits.HDUList([datahdu, freqhdu, timehdu])
        hdulist.writeto(savefile, overwrite=True)
        return






#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to read NenuFAR BHR data
        Originally by C. Viou
        Adapted to Nenupy by A. Loh
"""

__author__ = ['Alan Loh', 'Cedric Viou']
__copyright__ = 'Copyright 2019, nenupy'
__credits__ = ['Alan Loh', 'Cedric Viou']
__license__ = 'MIT'
__version__ = '0.1.0'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'Production'
__all__ = ['BHR']


import os
import numpy as np

from astropy import units as u
from astropy.time import Time

# ============================================================= #
# ---------------------------- BHR ---------------------------- #
# ============================================================= #

class BHR(object):
    """
    """
    def __init__(self, obsfile):
        self.obsfile = obsfile


    # ========================================================= #
    # --------------------- Getter/Setter --------------------- #
    @property
    def obsfile(self):
        """ BHR observaiton file, should end with '.spectra'
        """
        return self._obsfile
    @obsfile.setter
    def obsfile(self, o):
        assert o.endswith('.spectra'), 'Wrong file type'
        o = os.path.abspath(o)
        assert os.path.isfile(o), 'File {} not found'.format(o)

        head_dtype = np.dtype([('idx', 'uint64'),
                          ('TIMESTAMP', 'uint64'),
                          ('BLOCKSEQNUMBER', 'uint64'),
                          ('fftlen', 'int32'),
                          ('nfft2int', 'int32'),
                          ('fftovlp', 'int32'),
                          ('apodisation', 'int32'),
                          ('nffte', 'int32'),
                          ('nbchan', 'int32')])
        header = np.fromfile(o, count=1, dtype=head_dtype)[0]
        # self.lane = header['idx']
        self.nchan = header['nbchan']
        self.nffte = header['nffte']
        self.fftlen = header['fftlen']
        self.freq_res = 0.1953125 / self.fftlen * u.MHz
        self.time_res = self.fftlen * header['nfft2int'] * 5.12 * u.us

        chan_dtype = np.dtype([('lane', 'int32'),
                               ('beam', 'int32'),
                               ('channel', 'int32'),
                               ('fft0', 'float32', (header['nffte'], header['fftlen'], 2)),
                               ('fft1', 'float32', (header['nffte'], header['fftlen'], 2)),
                               ])

        block_dtype = np.dtype([('idx', 'uint64'),
                             ('TIMESTAMP', 'uint64'),
                             ('BLOCKSEQNUMBER', 'uint64'),
                             ('fftlen', 'int32'),
                             ('nfft2int', 'int32'),
                             ('fftovlp', 'int32'),
                             ('apodisation', 'int32'),
                             ('nffte', 'int32'),
                             ('nbchan', 'int32'),
                             ('data', chan_dtype, (header['nbchan'])),
                             ])

        self.alldata = np.fromfile(o, count=-1, dtype=block_dtype)



    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def select(self):
        """
        """
        max_bsn = 200e6 / 1024
        self.times = self.alldata['TIMESTAMP'] + self.alldata['BLOCKSEQNUMBER']/max_bsn
        self.times = Time(self.times, format='unix')

        # data_selected1 = b.alldata['data']['fft0'][0, :, :, 1, 0] # block number, nbchan, nffte, fftlen, polar
        # data_selected2 = b.alldata['data']['fft0'][1, :, :, 1, 0]
        # return np.hstack((data_selected1, data_selected2))

        # data_selected = b.alldata['data']['fft0'][0:3, :, :, :, 0]
        # nblocks, nchan, nt, fftl = data_selected.shape
        # data_selected = data_selected.transpose((0, 2, 1, 3)).copy()
        # data = data_selected.reshape((nblocks * self.nffte,
        #                      nchan * self.fftlen))

        data_selected = b.alldata['data']['fft0'][0:80, :, :, 0, 0]
        nblocks, nchan, nt = data_selected.shape
        data_selected = data_selected.transpose((0, 2, 1))
        data = data_selected.reshape((nblocks * self.nffte, nchan))
        return data

# ============================================================= #
# ============================================================= #

import pylab as plt
b = BHR(obsfile='/Users/aloh/Documents/Work/NenuFAR/Nenupy3_Tests/BHR/B1919+21_B1_TRACKING_BHR_20190212_123033_0.spectra')
# print(b.time_res, b.freq_res)
data = b.select()
plt.imshow(np.log10(data.T), origin='lower', aspect='auto')
plt.show()

















"""

Description of .spectra files as produced by 'tf' code with undysputed           (Feb 11, 2019)
/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
for each block
 ULONG    first effective sample index as incremented by the 'tf' code
            used for this block (starting with 0!)
 ULONG    TIMESTAMP of first sample (see NenuFAR/LOFAR beamlet description)
 ULONG    BLOCKSEQNUMBER of first sample (see NenuFAR/LOFAR beamlet description)
 LONGINT  fftlen : FFT length -> freq resolution= 0.1953125/fftlen MHz
 LONGINT  nfft2int : Nber of integrated FFTs -> time resolution= fftlen*nfft2int*5.12 us
 LONGINT  ffovlp : FFToverlap 0/1                    (not used yet)
 LONGINT  apodisation : (0: none, 54=hamming, ...)   (not used yet)
 LONGINT  nffte : nber of FFTs to be read per beamlet within this block
             (so we have 'nffte' spectra/time_sample per block)
 LONGINT  nbchan : nbr of subsequent beamlets/channels (nb of NenuFAR/LOFAR 195kHz beamlets)
             (so we have 'fftlen*nbchan' frequencies for each spectra)

 for each beamlet
  LONGINT  lane index (0, 1, 2 or 3)
  LONGINT  beam index (numerical beam : 0, 1, ...)
  LONGINT  beamlet/channel index (0...768, 0 is 0.000MHz)
  FLOAT    pol0 data : 'nffte' time_samples each made of 'fftlen' pairs [XrXr+XiXi : YrYr+YiYi]
  FLOAT    pol1 data : 'nffte' time_samples each made of 'fftlen' pairs [XrYr+XiYi : XrYi-XiYr]
/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/


Calculs de Stokes (booklog du 6 janvier 2009)

LINEAR FEEDS
  les polars complexes sont (X,Y) ou (Xr,Xi,Yr,Yi) pour (E,W) ou (H,V)

ASP files
  dans les fichiers .asp on trouve
       XX = XrXr + XiXi
       YY = YrYr + YiYi
       Re(X*Y) = XrYr + XiYi
       Im(X*Y) = XrYi - XiYr

ASC files
  in .asc files without "-S" option of asp_prof2Dn (added on Nov 16, 2012)
       XX = XrXr + XiXi
       YY = YrYr + YiYi
       Re(X*Y) = XrYr + XiYi
       Im(X*Y) = XrYi - XiYr

ASC files
  dans les fichiers .asc avec option "-S" de asp_prof, asp_prof2Dn
      I = (XX+YY)/2
      Q = (XX-YY)/2
      U = Re(X*Y)
      V = Im(X*Y)           voir p121 Handbook of Pulsar Astronomy

dans les plots/figures avec l'option 'pol' de pltasc
      IntTot = I
      Circ = V
      Lin = sqrt(Q*Q+U*U)
      PA = 1/2 ArcTg(U/Q)   voir p187 Handbook of Pulsar Astronomy

Added on Nov 23, 2014

CIRCULAR FEEDS
  les polars complexes sont (R,L) ou (Rr,Ri,Lr,Li), alors
     RR = RrRr + RiRi
     LL = LrLr + LiLi
     Re(R*.L) = RrLr + RiLi
     Im(R*.L) = RrLi - RiLr
  et
     I = (RR+LL)/2
     Q = Re(R*.L)
     U = Im(R*.L)
     V = (RR-LL)/2
  si (X,Y) permettent de calculer XX, YY, Re(X*Y) et Im(X*Y),
 alors on peut retrouver RR et LL avec :
   RR = (XX+YY)/2 + Im(X*Y)
   LL = (XX+YY)/2 - Im(X*Y)


"""





#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to handle antenna models
        by A. Loh
"""

import os
import sys
import numpy as np

from scipy.interpolate import interp2d

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['AntennaModel']

class AntennaModel():
    """ This class returns antenna models
        The outputs are response functions: antenna_gain(azim, elev)
        Parameters:
        - design (str): design of the antenna
        - kwargs: keywords used for specific antenna models
    """
    def __init__(self, design='nenufar', **kwargs):
        self.kwargs = kwargs
        self.design = design
        

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def design(self):
        """ Antenna design
        """
        return self._design
    @design.setter
    def design(self, d):
        if isinstance(d, np.ndarray):
            self._makeCustom(model=d)
        elif d.lower() == 'nenufar':
            self._design = 'nenufar'
            self._getNenuFAR()
        else:
            raise Exception("\n\t=== Nothing handled yet ===")
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def plotAntGain(self):
        """
        """
        from matplotlib import pyplot as plt
        gain = self.antenna_gain(np.linspace(0,360,360), np.linspace(0,90,90))
        plt.imshow(gain, origin='lower')
        plt.show()
        return

    def _makeCustom(self, model):
        gaininterp = interp2d( np.linspace(0., 360., model.shape[1]), 
             np.linspace(0., 90., model.shape[0]), model, kind='linear' )
        self.antenna_gain = gaininterp
        self.antenna_freq = self.kwargs['freq']
        return
    
    def _getNenuFAR(self):
        """
        """
        if 'freq' not in self.kwargs.keys():
            raise Exception("\n\t=== keyword 'freq' needs to be defined  ===")
        if 'polar' not in self.kwargs.keys():
            raise Exception("\n\t=== keyword 'polar' needs to be defined  ===")
        
        from scipy.io.idl import readsav
        from scipy.interpolate import interp2d
        
        antpol = {'NE': 'NE_SW', 'NW': 'NW_SE'}
        pol    = antpol[self.kwargs['polar']]
        f1 = int( np.floor( self.kwargs['freq']/10. ) ) * 10
        f2 = int( np.ceil(  self.kwargs['freq']/10. ) ) * 10

        modulepath = os.path.dirname( os.path.realpath(__file__) )

        gain_inf = 10**( readsav( os.path.join(modulepath, 'LSS_GRID_AG_'+pol+'_'+str(f1)+'_astro.sav') )['ga']/10. )
        gain_sup = 10**( readsav( os.path.join(modulepath, 'LSS_GRID_AG_'+pol+'_'+str(f2)+'_astro.sav') )['ga']/10. )
        if f1 != f2:
            gain = gain_inf * (f2-self.kwargs['freq']) / 10. + gain_sup * (self.kwargs['freq']-f1) / 10.
        else:
            gain = gain_inf

        gaininterp = interp2d( np.linspace(0., 360., gain.shape[1]), 
             np.linspace(0., 90., gain.shape[0]), gain, kind='linear' )
        
        self.antenna_gain = gaininterp
        self.antenna_freq = self.kwargs['freq']

        return





#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to compute a SST NenuFAR beam
        by A. Loh
"""

import os
import sys
import numpy as np

import matplotlib as mpl
from matplotlib import pyplot as plt
import matplotlib.ticker as mtick
from scipy.io.idl import readsav
from scipy.interpolate import interp1d
from astropy.io import fits

import healpy as hp
from astropy import constants as const

from . import PhasedArrayBeam
from .antenna import AntennaModel
from .antenna import miniarrays
from ..read import SST

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['SSTbeam', 'SSTbeamHPX']

class SSTbeam(object):
    """ 
    Class to handle SST beams
    
    Parameters
    ----------
    * sst : `SST` class, optional
        Instance of *SST class*, it refers to a particular observation
    * ma : integer, optional
        Mini-Array number (automatic by default)
    * marotation : float, optional
        Mini-Array rotation (automatic by default)
    * polar : str, optional
        Polarization, 'NE' or 'NW' ('NW' by default)
    * freq : float, optional
        Frequency in MHz (50 by default)
    * azana : float, optional
        Azimuth in degrees (180 by default)
    * elana : float, optional
        Elevation in degrees (90 by default)

    Returns
    -------
    beam : np.ndarray
        Normalized 2D array of shape (azimuth, elevation)

    Examples
    --------
    To plot a beam:
    
    >>> from nenupy.beam import SSTbeam
    >>> beam = SSTbeam(ma=12, azana=30, elana=45, freq=65, polar='NE')
    >>> beam.plotBeam()

    Work on the beam computed from a given SST observation:

    >>> from nenupy.read import SST
    >>> from nenupy.beam import SSTbeam
    >>> sst = SST('observation_SST.fits')
    >>> sst.freq = 45
    >>> sst.polar = 'ne'
    >>> beam = SSTbeam(sst)
    >>> beam.getBeam()
    """
    def __init__(self, sst=None, **kwargs):
        self.sst = sst
        self._evalkwargs(kwargs)
        
    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def sst(self):
        """ SST observation
        """
        return self._sst
    @sst.setter
    def sst(self, s):
        if isinstance(s, SST):
            self._sst       = s
            self.ma         = self._sst.ma
            self.marotation = self._sst.marotation
            self.freq       = self._sst.freq
            self.polar      = self._sst.polar
            self.azana      = self._sst.azana 
            self.elana      = self._sst.elana
        else:
            self._sst = s
        return

    @property
    def freq(self):
        """ Frequency selection in MHz
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if not isinstance(f, (int, float, np.float32, np.float64)):
            raise TypeError("\n\t=== Attribute 'freq' should be a number ===")
        elif (f<=0.) or (f>100.):
            raise ValueError("\n\t=== Attribute 'freq' set to a value outside of NenuFAR's frequency range ===")
        else:
            self._freq = f
            return

    @property
    def polar(self):
        """ Polarization selection
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, (str)):
            raise TypeError("\n\t=== Attribute 'polar' should be a string ===")
        elif p.upper() not in ['NW', 'NE']:
            raise ValueError("\n\t=== Attribute 'polar' does not correspond to either 'NW' or 'NE' ===")
        else:
            self._polar = p.upper()
            return

    @property
    def ma(self):
        """ Miniarray selection
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        if self.sst is None:
            marecorded = miniarrays.ma[:, 0].astype(int)
        else:
            marecorded = self.sst.miniarrays
        if m is None:
            print("\n\t==== WARNING: ma is set by default ===")
            m = marecorded[0]
        if not isinstance(m, (int, np.integer) ):
            raise TypeError("\n\t=== Attribute 'ma' should be an integer ===")
        elif not m in marecorded:
            raise ValueError("\n\t=== Attribute 'ma' contains miniarray index not matching existing MAs (up to {}) ===".format(marecorded.max()))
        else:
            self._ma = m
            return

    @property
    def marotation(self):
        """ MA rotation selection
        """
        #if (self.sst is None) or (
        if isinstance(self._marotation, np.ndarray):
            return self._marotation[self.ma]
        else:
            return self._marotation
    @marotation.setter
    def marotation(self, r):
        if r is None:
            print("\n\t==== WARNING: MA rotation is set by default ===")
            self._marotation = miniarrays.ma[:, 1]
            return
        if not isinstance(r, (float, int, np.float32, np.float64, np.int16, np.int32, np.int64)):
            raise TypeError("\n\t=== Attribute 'marotation' should be a number ===")
        else:
            self._marotation = r
            return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getBeam(self):
        """ Get the SST beam
        """
        model = AntennaModel(design='nenufar', freq=self.freq, polar=self.polar)
        elevation = self._squintMA(self.elana)
        az, el = self._realPointing(self.azana, elevation)
        # # az   -= self.marotation + 90. # origin correction
        # az   -= 90.
        beam  = PhasedArrayBeam(p=self._antPos(), m=model, a=az, e=el)
        self.beam = beam.getBeam()
        # self.beam = np.roll(self.beam, int(1./beam.resol)*int(self.marotation + 90) , axis=1)
        # self.beam = np.roll(self.beam, int(1./beam.resol)*int(90) , axis=1)
        return

    def plot(self, store=None, **kwargs):
        """ Plot the SST Beam
        """
        self.getBeam()

        theta = np.linspace(0., 90., self.beam.shape[0])
        phi   = np.radians( np.linspace(0., 360., self.beam.shape[1]) )
        # ------ Plot ------ #
        fig = plt.figure(figsize=(18/2.54, 18/2.54))
        ax  = fig.add_subplot(111, projection='polar')
        normcb = mpl.colors.LogNorm(vmin=self.beam.max() * 1.e-4, vmax=self.beam.max())
        p = ax.pcolormesh(phi, theta, self.beam, norm=normcb, **kwargs)
        ax.grid(linestyle='-', linewidth=0.5, color='white', alpha=0.4)
        plt.setp(ax.get_yticklabels(), rotation='horizontal', color='white')
        g = lambda x,y: r'%d'%(90-x)
        ax.yaxis.set_major_formatter(mtick.FuncFormatter( g ))
        if store is not None:
            plt.savefig(store, dpi=300, facecolor='none',
                edgecolor='none', transparent=True, bbox_inches='tight')
        else:
            plt.title('MA={}, pol={}, freq={:.2f}MHz, az={}, el={}'.format(self.ma, self.polar, self.freq, self.azana, self.elana))
            plt.show()
        plt.close('all')
        return

    def save(self, savefile=None):
        """ Save the beam
        """
        if savefile is None:
            savefile = 'beam.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        self.getBeam()

        prihdr = fits.Header()
        prihdr.set('FREQ', str(self.freq))
        prihdr.set('POLAR', self.polar)
        prihdr.set('MINI-ARR', str(self.ma))
        prihdr.set('MA-ROT', str(self.marotation))
        datahdu = fits.PrimaryHDU( np.fliplr(self.beam).T, header=prihdr)
        hdulist = fits.HDUList([datahdu])
        hdulist.writeto(savefile, overwrite=True)
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _antPos(self):
        """ Return the antenna position within a mini-array
        """
        antpos = miniarrays.antpos

        if self.marotation is not None:
            rot = np.radians( self.marotation )
            rotation = np.array([[ np.cos(rot),  np.sin(rot), 0],
                                 [-np.sin(rot),  np.cos(rot), 0],
                                 [ 0,            0,           1]])
            antpos = np.dot( antpos, rotation )
        return antpos

    def _realPointing(self, a, e):
        """ Find the real pointing direction from self.ra, self.dec
            Nenufar can only point towards directions on a 128 x 128 cells grid
            Therefore PZ computed for each -desired- direction at a resolution of 0.05deg
            the corresponding real observed direction.
        """ 
        modulepath = os.path.dirname( os.path.realpath(__file__) )
        thph = fits.getdata( os.path.join(modulepath, 'NenuFAR_thph.fits') )
        theta, phi = ( int((90.-e)/0.05 - 0.5), int(a/0.05 - 0.5)  )
        t, p = thph[:, theta, phi]
        return p, 90.-t

    def _squintMA(self, e):
        """ Compute the elevation to be pointed that take into account the squint if we want to observe ele
        """
        modulepath = os.path.dirname( os.path.realpath(__file__) )
        squint  = readsav( os.path.join(modulepath, 'squint_table.sav') )
        optfreq = 30
        indfreq = np.where(squint['freq']==optfreq)[0][0]
        newele  = interp1d(squint['elev_desiree'][indfreq,:], squint['elev_a_pointer'])(e)
        if newele < 20.: # squint is limited at 20 deg elevation
            newele = 20.
        return newele

    def _evalkwargs(self, kwargs):
        """
            Parameters
            ----------
            kwargs : dictionnary
                Dictionnary of keys and values to look from in order to fill the class attributes
        """
        allowed = ['ma', 'marotation', 'freq', 'polar', 'azana', 'elana']
        if not all([ki in allowed for ki in kwargs.keys()]):
            print("\n\t=== WARNING: unkwnown keywords, authorized: {} ===".format(allowed))
        if not hasattr(self, 'ma'):
            self.ma         = kwargs.get('ma', None)
        if not hasattr(self, 'marotation'):
            self.marotation = kwargs.get('marotation', None)
        if not hasattr(self, 'freq'):
            self.freq       = kwargs.get('freq', 50)
        if not hasattr(self, 'polar'):
            self.polar      = kwargs.get('polar', 'NW')
        if not hasattr(self, 'azana'):
            self.azana      = kwargs.get('azana', 180.)
        if not hasattr(self, 'elana'):
            self.elana      = kwargs.get('elana', 90.)
        return





















# ============================================================= #
# ============================================================= #
class SSTbeamHPX(object):
    """
    """

    def __init__(self, sst=None, **kwargs):
        self.sst = sst
        self._evalkwargs(kwargs)

    # ========================================================= #
    # ==================== Getter / Setter ==================== #
    @property
    def sst(self):
        """ SST observation
        """
        return self._sst
    @sst.setter
    def sst(self, s):
        if isinstance(s, SST):
            self._sst       = s
            self.ma         = self._sst.ma
            self.marotation = self._sst.marotation
            self.freq       = self._sst.freq
            self.polar      = self._sst.polar
            self.azana      = self._sst.azana 
            self.elana      = self._sst.elana
        else:
            self._sst = s
        return

    # --------------------------------------------------------- #
    @property
    def azana(self):
        """ Pointed azimuth in degrees
        """
        return self._azana
    @azana.setter
    def azana(self, a):
        self._azana = np.radians(a)
        return

    # --------------------------------------------------------- #
    @property
    def elana(self):
        """ Pointed elevation in degrees
        """
        return self._elana
    @elana.setter
    def elana(self, e):
        self._elana = np.radians(e)
        return

    # --------------------------------------------------------- #
    @property
    def freq(self):
        """ Frequency selection in MHz
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if not isinstance(f, (int, float, np.float32, np.float64)):
            raise TypeError("\n\t=== Attribute 'freq' should be a number ===")
        elif (f<=0.) or (f>100.):
            raise ValueError("\n\t=== Attribute 'freq' set to a value outside of NenuFAR's frequency range ===")
        else:
            self._freq = f
            self.wavevec = 2 * np.pi * f * 1.e6 / const.c.value
            return

    # --------------------------------------------------------- #
    @property
    def polar(self):
        """ Polarization selection
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, (str)):
            raise TypeError("\n\t=== Attribute 'polar' should be a string ===")
        elif p.upper() not in ['NW', 'NE']:
            raise ValueError("\n\t=== Attribute 'polar' does not correspond to either 'NW' or 'NE' ===")
        else:
            self._polar = p.upper()
            return

    # --------------------------------------------------------- #
    @property
    def ma(self):
        """ Miniarray selection
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        if self.sst is None:
            marecorded = miniarrays.ma[:, 0].astype(int)
        else:
            marecorded = self.sst.miniarrays
        if m is None:
            # print("\n\t==== WARNING: ma is set by default ===")
            m = marecorded[0]
        if not isinstance(m, (int, np.integer) ):
            raise TypeError("\n\t=== Attribute 'ma' should be an integer ===")
        elif not m in marecorded:
            raise ValueError("\n\t=== Attribute 'ma' contains miniarray index not matching existing MAs (up to {}) ===".format(marecorded.max()))
        else:
            self._ma = m
            return

    # --------------------------------------------------------- #
    @property
    def marotation(self):
        """ MA rotation selection
        """
        #if (self.sst is None) or (
        if isinstance(self._marotation, np.ndarray):
            return self._marotation[self.ma]
        else:
            return self._marotation
    @marotation.setter
    def marotation(self, r):
        if r is None:
            # print("\n\t==== WARNING: MA rotation is set by default ===")
            self._marotation = miniarrays.ma[:, 1]
            return
        if not isinstance(r, (float, int, np.float32, np.float64, np.int16, np.int32, np.int64)):
            raise TypeError("\n\t=== Attribute 'marotation' should be a number ===")
        else:
            self._marotation = r
            return

    @property
    def resolution(self):
        """ Angular resolution in degrees
        """
        return self._resolution
    @resolution.setter
    def resolution(self, r):
        # Find the nside corresponding to the desired resolution
        nsides = np.array([2**i for i in range(1, 12)])
        resol_deg = np.degrees(hp.nside2resol(nside=nsides, arcmin=False))
        idx = (np.abs(resol_deg - r)).argmin()
        self._nside = nsides[idx]
        self._npix = hp.nside2npix(self._nside)        
        self._resolution = resol_deg[idx]

        self._sky_grid()
        return

    # ========================================================= #
    def compute(self):
        """ Compute the phased array beam
        """
        # ------ Phase delay towards pointing ------ #
        elevation = self._squint(self.elana)
        az, el = self._real_pointing(self.azana, elevation)
        ux = np.cos(az) * np.cos(el)
        uy = np.sin(az) * np.cos(el)
        uz = np.cos(el)
        antpos = self._ant_pos()
        phix = antpos[:, 0] * ux[np.newaxis]
        phiy = antpos[:, 1] * uy[np.newaxis]
        phiz = antpos[:, 2] * uz[np.newaxis]
        phi  = self.wavevec * (phix + phiy + phiz)

        # ------ Phase reference ------ #
        _xp = np.cos(self._az) * np.cos(self._alt) 
        _yp = np.sin(self._az) * np.cos(self._alt) 
        _zp = np.sin(self._alt)
        phi0x = antpos[:, 0] * _xp[:, np.newaxis]
        phi0y = antpos[:, 1] * _yp[:, np.newaxis]
        phi0z = antpos[:, 2] * _zp[:, np.newaxis]
        phi0  = self.wavevec * (phi0x + phi0y + phi0z)
        
        phase = phi0[:, :] - phi[np.newaxis, :]

        # ------ e^(i Phi) ------ #
        eiphi = np.sum( np.exp(1j * phase), axis=1 )
        self.beam = np.real(eiphi * eiphi.conjugate())
        self.beam[self.below_horizon] = 0.
        self.beam *= self._nenufar_antenna()
 
        return

    # --------------------------------------------------------- #
    def plot(self):
        """
        """
        if not hasattr(self, 'beam'):
            self.compute()
        hp.orthview( np.log10(self.beam), rot=[0, 90, 90])
        hp.graticule()
        plt.show()
        plt.close('all')
        return

    # ========================================================= #
    # ========================================================= #
    def _sky_grid(self):
        """ Compute the sky grid
        """
        self._az, self._alt = np.radians(hp.pix2ang(nside=self._nside,
                                            ipix=np.arange(self._npix),
                                            lonlat=True,
                                            nest=False))
        self._az = self._az[::-1] # east is 90        
        # Remove local altitudes < 0deg
        self.below_horizon = (self._alt < 0.)
        return

    # --------------------------------------------------------- #
    def _ant_pos(self):
        """ Return the antenna position within a mini-array
        """
        antpos = miniarrays.antpos
        if self.marotation is not None:
            rot = np.radians( self.marotation )
            rotation = np.array([[ np.cos(rot), np.sin(rot), 0],
                                 [-np.sin(rot), np.cos(rot), 0],
                                 [ 0,           0,           1]])
            antpos = np.dot( antpos, rotation )
        return antpos

    def _nenufar_antenna(self):
        """ NenuFAR antenna gain
        """
        from scipy.interpolate import interp2d

        antpol = {'NE': 'NE_SW', 'NW': 'NW_SE'}
        pol    = antpol[self.polar]
        f1 = int( np.floor( self.freq/10. ) ) * 10
        f2 = int( np.ceil(  self.freq/10. ) ) * 10

        modulepath = os.path.dirname( os.path.realpath(__file__) )

        gain_inf = 10**( readsav( os.path.join(modulepath, 'antenna/LSS_GRID_AG_'+pol+'_'+str(f1)+'_astro.sav') )['ga']/10. )
        gain_sup = 10**( readsav( os.path.join(modulepath, 'antenna/LSS_GRID_AG_'+pol+'_'+str(f2)+'_astro.sav') )['ga']/10. )
        if f1 != f2:
            gain = gain_inf * (f2-self.kwargs['freq']) / 10. + gain_sup * (self.kwargs['freq']-f1) / 10.
        else:
            gain = gain_inf

        gaininterp = interp2d( np.linspace(0., 360., gain.shape[1]), 
             np.linspace(0., 90., gain.shape[0]), gain, kind='linear' )

        ant_gain = np.squeeze(list(map(gaininterp, np.degrees(self._az), np.degrees(self._alt))))

        # # WRITE THE FILE
        # import os
        # import numpy as np
        # import healpy as hp
        # from scipy.io.idl import readsav
        # from scipy.interpolate import interp2d
        # module = '/Users/aloh/Documents/Work/Scripts/Own_Packages/nenupy/nenupy/beam/antenna'
        # azgrid, elgrid = hp.pix2ang(nside=64,#1024,
        #     ipix=np.arange(hp.nside2npix(64)),
        #     lonlat=True,
        #     nest=False)
        # azgrid = np.radians(azgrid[::-1]) # east is 90
        # elgrid = np.radians(elgrid)
        # ant_gain = []
        # col_name = []
        # for pol in ['NW_SE', 'NE_SW']:
        #     for freq in [10, 20, 30, 40, 50, 60, 70, 80]:
        #         gain = 10**( readsav( os.path.join(module, 'LSS_GRID_AG_{}_{}_astro.sav'.format(pol, freq)) )['ga']/10.)
        #         gaininterp = interp2d( np.linspace(0., 360., gain.shape[1]), 
        #              np.linspace(0., 90., gain.shape[0]), gain, kind='linear' )
        #         ant_gain.append( np.squeeze(list(map(gaininterp, np.degrees(azgrid), np.degrees(elgrid)))) )
        #         col_name.append( '{}_{}'.format(pol[:2], freq) )
        # hp.write_map(filename="NenuFAR_Ant_Hpx.fits",
        #     m=ant_gain,
        #     column_names=col_name)
        
        return ant_gain/ant_gain.max()

    # --------------------------------------------------------- #
    def _real_pointing(self, a, e):
        """ Find the real pointing direction from self.ra, self.dec
            Nenufar can only point towards directions on a 128 x 128 cells grid
            Therefore PZ computed for each -desired- direction at a resolution of 0.05deg
            the corresponding real observed direction.
        """ 
        modulepath = os.path.dirname( os.path.realpath(__file__) )
        thph = fits.getdata( os.path.join(modulepath, 'NenuFAR_thph.fits') )
        theta, phi = ( int((90.-np.degrees(e))/0.05 - 0.5), int(np.degrees(a)/0.05 - 0.5)  )
        t, p = thph[:, theta, phi]
        return np.radians(p), np.radians(90.-t)

    # --------------------------------------------------------- #
    def _squint(self, e):
        """ Compute the elevation to be pointed that take into account
            the squint if we want to observe ele
        """
        modulepath = os.path.dirname( os.path.realpath(__file__) )
        squint  = readsav( os.path.join(modulepath, 'squint_table.sav') )
        optfreq = 30
        indfreq = np.where(squint['freq']==optfreq)[0][0]
        newele  = interp1d(squint['elev_desiree'][indfreq,:], squint['elev_a_pointer'])(np.degrees(e))
        if newele < 20.: # squint is limited at 20 deg elevation
            newele = 20.
        return np.radians(newele)

    # --------------------------------------------------------- #
    def _evalkwargs(self, kwargs):
        """
            Parameters
            ----------
            kwargs : dictionnary
                Dictionnary of keys and values to look from in order to fill the class attributes
        """
        allowed = ['ma', 'marotation', 'freq', 'polar', 'azana', 'elana', 'resolution']
        if not all([ki in allowed for ki in kwargs.keys()]):
            print("\n\t=== WARNING: unkwnown keywords, authorized: {} ===".format(allowed))
        if not hasattr(self, 'ma'):
            self.ma         = kwargs.get('ma', None)
        if not hasattr(self, 'marotation'):
            self.marotation = kwargs.get('marotation', None)
        if not hasattr(self, 'freq'):
            self.freq       = kwargs.get('freq', 50)
        if not hasattr(self, 'polar'):
            self.polar      = kwargs.get('polar', 'NW')
        if not hasattr(self, 'azana'):
            self.azana      = kwargs.get('azana', 180.)
        if not hasattr(self, 'elana'):
            self.elana      = kwargs.get('elana', 90.)
        if not hasattr(self, 'resolution'):
            self.resolution = kwargs.get('resolution', 0.2)
        return




#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
    Examples
    --------
        - Transit prediction (whole day, zenith)
        ```
        from nenupy.hpx import Transit
        t=Transit()
        t.predict(src=None, resol=1, freq=55, duration=86400, miniarrays=[10], azana=180, elana=90, azdig=180, eldig=90)
        ```

        - Transit prediction with a pointsource model
        ```
        from nenupy.hpx import Transit
        t=Transit()
        t.predict(src='Cyg A', resol=0.2, freq=60, duration=3600, skymodel='pointsource', ra=299.868, dec=40.734)
        ``` 

        - Tracking prediction of Cyg A using the GSM:
        ```
        from nenupy.hpx import Tracking
        t=Tracking()
        t.predict(src='Cyg A', duration=14400, dt=600, miniarrays=0, resol=1, freq=60)
        ```

        - Tracking prediction of Cyg A using a pointsource skymodel:
        ```
        from nenupy.hpx import Tracking
        t=Tracking()
        t.predict(src='Cyg A', time='2019-03-05 08:00:03', duration=7200, dt=600, miniarrays=0, resol=1, freq=60, skymodel='pointsource', ra=299.868, dec=40.734)
        ```
"""

# import os, sys
import numpy as np
import healpy as hp
from astropy.time import Time, TimeDelta

from nenupy.read import BST
from nenupy.hpx import Anabeam, Digibeam, Skymodel
from nenupy.astro import getTransit, getAltaz, getTime, getSrc
from nenupy.utils import ProgressBar

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2019, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['Simu', 'Transit', 'Tracking']




# ============================================================= #
# ---------------------------- Simu --------------------------- #
# ============================================================= #
class Simu(object):
    """
    """
    def __init__(self, **kwargs):
        self.dt = 60
        self.azana = 180
        self.elana = 90
        self.azdig = 180
        self.eldig = 90
        self.nomas = None
        self.freq  = 50
        self.polar = 'NW'
        self.resol = 0.2
        self.skymodel = 'gsm'
        self._simukwargs(kwargs)

    # ========================================================= #
    # --------------------- Getter/Setter --------------------- #
    @property
    def dt(self):
        """ Simulation time resolution in seconds
        """
        return self._dt
    @dt.setter
    def dt(self, d):
        if not isinstance(d, TimeDelta):
            try:
                d = TimeDelta(d, format='sec')
            except:
                raise AttributeError("\n\t=== 'dt' format not understood ===")
        self._dt = d
        return
    # --------------------------------------------------------- #
    @property
    def resol(self):
        """ Angular resolution in degrees
        """
        return self._resol
    @resol.setter
    def resol(self, r):
        if r < 0.1:
            print('WARNING, high-res (<0.1deg) may result in long computational durations')
        # Find the nside corresponding to the desired resolution
        nsides = np.array([2**i for i in range(1, 12)])
        resol_rad = hp.nside2resol(nside=nsides, arcmin=False)
        resol_deg = np.degrees(resol_rad)
        idx = (np.abs(resol_deg - r)).argmin()
        if hasattr(self, 'resol'):
            if resol_deg[idx] == self.resol:
                # No need to recompute it
                return
        self._resol = resol_deg[idx]
        self.nside = nsides[idx]
        return
    # --------------------------------------------------------- #
    @property
    def model(self):
        """ Skymodel object
        """
        m = Skymodel(nside=self.nside,
                     freq=self.freq)
        return m

    # ========================================================= #
    # ----------------------- Internal ------------------------ #
    def _compute_beam(self):
        """ Compute a numeric beam
        """
        self.db = Digibeam(azana=self.azana,
            azdig=self.azdig,
            elana=self.elana,
            eldig=self.eldig,
            miniarrays=self.nomas,
            freq=self.freq,
            polar=self.polar,
            resol=self.resol)
        self.beam = self.db.get_digibeam()
        return
    # --------------------------------------------------------- #
    def _transit_sky(self, **kwargs):
        """ Perform a sky rotation from `self.start` to `self.stop`
            with `dt` sec steps.
            Compute the -static- beam and integrate everything.
        """
        self.simulation = {'time': [],
                            'amp': []}
        sm = self.model
        
        nbtime  = int(np.ceil( (self.stop - self.start).sec / self.dt.sec ))
        self.dt = (self.stop - self.start) / (nbtime - 1)
        
        bar = ProgressBar(valmax=nbtime, title='Transit observation simulation')
        for i in range(nbtime):
            current = self.start + i*self.dt
            # Sky rotation:
            sky = sm.get_skymodel(time=current, model=self.skymodel, **kwargs)
            # Integrate the beam x sky
            integ = np.sum(self.beam * sky )#* self.db.domega ) #/ (4 * np.pi) #self.db.domega ) no need cause hpx?

            self.simulation['time'].append(current.mjd)
            self.simulation['amp'].append(integ)            
            bar.update()

        self.simulation['time'] = Time(np.array(self.simulation['time']),
            format='mjd')
        self.simulation['amp'] = np.array(self.simulation['amp'])
        return
    # --------------------------------------------------------- #
    def _tracking_sky(self, src=None, obs=None, **kwargs):
        """ Perform a sky rotation from `self.start` to `self.stop`
            with `dt` sec steps.
            For each time step, a new beam is computed toward `src`.
        """
        if obs is None:
            src = getSrc(src)

        self.simulation = {'time': [],
                            'amp': []}
        sm = self.model
        
        nbtime  = int(np.ceil( (self.stop - self.start).sec / self.dt.sec ))
        self.dt = (self.stop - self.start) / (nbtime - 1)

        bar = ProgressBar(valmax=nbtime, title='Tracking observation simulation')
        for i in range(nbtime):
            current = self.start + i*self.dt
            # Sky rotation:
            sky = sm.get_skymodel(time=current, model=self.skymodel, **kwargs)
            # Compute the beam
            if obs is None:
                altaz = getAltaz(source=src,
                    time=current,
                    loc='NenuFAR')
                self.azana = altaz.az.deg
                self.elana = altaz.alt.deg
                self.azdig = altaz.az.deg
                self.eldig = altaz.alt.deg
            else:
                obs.time = current
                self.azana = obs.azana
                self.elana = obs.elana
                self.azdig = obs.azdig
                self.eldig = obs.eldig
            self._compute_beam()
            # Integrate the beam x sky
            integ = np.sum(self.beam * sky) # / (4 * np.pi) # * self.db.domega)

            self.simulation['time'].append(current.mjd)
            self.simulation['amp'].append(integ)            
            bar.update()

        self.simulation['time'] = Time(np.array(self.simulation['time']),
            format='mjd')
        self.simulation['amp'] = np.array(self.simulation['amp'])
        return
    # --------------------------------------------------------- #
    def _simukwargs(self, kwargs):
        """
        """
        for key, value in kwargs.items():
            if   key == 'polar': self.polar = value
            elif key == 'freq': self.freq = value
            elif key == 'time': self.time = value
            elif key == 'azana': self.azana = value
            elif key == 'elana': self.elana = value
            elif key == 'azdig': self.azdig = value
            elif key == 'eldig': self.eldig = value
            elif key == 'miniarrays': self.nomas = value
            elif key == 'resol': self.resol = value
            elif key == 'dt': self.dt = value
            elif key == 'skymodel': self.skymodel = value
            else:
                pass
        return
# ============================================================= #
# ============================================================= #




# ============================================================= #
# -------------------------- Transit -------------------------- #
# ============================================================= #
class Transit(Simu):
    """
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def predict(self, src=None, az=180, time='now', duration=1800, **kwargs):
        """ Compute the next source `src` transit at azimuth `az`
            from `time` for `duration` sec around the transit.

            src: str or tuple
                Name of the source or RA/Dec coordinates 
        """
        self._simukwargs(kwargs)

        duration = TimeDelta(duration, format='sec')

        if src is not None:
            self.t_transit = getTransit(source=src,
                time=time,
                loc='NenuFAR',
                az=az)
            altaz = getAltaz(source=src,
                time=self.t_transit,
                loc='NenuFAR')
            self.azana = altaz.az.deg
            self.elana = altaz.alt.deg
            self.azdig = altaz.az.deg
            self.eldig = altaz.alt.deg

            self.start = self.t_transit - duration/2.
            self.stop = self.t_transit + duration/2.
        else:
            time = getTime(time)
            self.start = time
            self.stop = time + duration

        self._compute_beam()

        self._transit_sky(**kwargs)
        return
    # --------------------------------------------------------- #
    def from_bst(self, obs, dbeam=0, **kwargs):
        """ Do the simulaiton based on an existing BST observation
        """
        self._simukwargs(kwargs)

        if not isinstance(obs, BST):
            bst = BST(obs)
        assert bst.type == 'transit',\
            'Obs is not a transit.'
        bst.select(freq=self.freq, polar=self.polar, dbeam=dbeam)

        self.azana = bst.azana
        self.elana = bst.elana
        self.azdig = bst.azdig
        self.eldig = bst.eldig
        self.nomas = bst.ma
        self.freq = bst.data['freq'][0]
        self.polar = bst.polar
        self._compute_beam()

        self.start = bst.data['time'][0]
        self.stop = bst.data['time'][-1]

        self._transit_sky(**kwargs)
        return
    # --------------------------------------------------------- #
    def plot(self):
        import pylab as plt
        x = (self.simulation['time'] - self.simulation['time'][0]).sec / 60
        if hasattr(self, 't_transit'):
            plt.axvline(x=(self.t_transit - self.simulation['time'][0]).sec/60,
                linestyle='--',
                linewidth=0.8,
                color='black')
        plt.plot(x, 10 * np.log10(self.simulation['amp']))
        plt.xlabel('Time (min since {})'.format(self.simulation['time'][0].iso))
        plt.ylabel('dB')
        plt.show()
        plt.close('all')
        return
# ============================================================= #
# ============================================================= #





# ============================================================= #
# ------------------------- Tracking -------------------------- #
# ============================================================= #
class Tracking(Simu):
    """
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # ========================================================= #
    # ------------------------ Methods ------------------------ #
    def predict(self, src=None, time='now', duration=1800, **kwargs):
        """ Compute the next source `src` transit at azimuth `az`
            from `time` for `duration` sec around the transit.

            src: str or tuple
                Name of the source or RA/Dec coordinates 
        """
        self._simukwargs(kwargs)

        duration = TimeDelta(duration, format='sec')
        self.start = getTime(time)
        self.stop = self.start + duration

        self._tracking_sky(src=src, **kwargs)
        return
    # --------------------------------------------------------- #
    def from_bst(self, obs, dbeam=0, **kwargs):
        """ Do the simulaiton based on an existing BST observation
        """
        self._simukwargs(kwargs)

        if not isinstance(obs, BST):
            bst = BST(obs)
        assert bst.type == 'tracking',\
            'Obs is not a tracking.'
        bst.select(freq=self.freq, polar=self.polar, dbeam=dbeam)

        self.azana = bst.azana
        self.elana = bst.elana
        self.azdig = bst.azdig
        self.eldig = bst.eldig
        self.nomas = bst.ma
        self.freq = bst.data['freq'][0]
        self.polar = bst.polar

        self.start = bst.data['time'][0]
        self.stop = bst.data['time'][-1]

        self._tracking_sky(obs=bst, **kwargs)
        return
    # --------------------------------------------------------- #
    def plot(self):
        import pylab as plt
        x = (self.simulation['time'] - self.simulation['time'][0]).sec / 60
        plt.plot(x, 10 * np.log10(self.simulation['amp']))
        plt.xlabel('Time (min since {})'.format(self.simulation['time'][0].iso))
        plt.ylabel('dB')
        plt.show()
        plt.close('all')
        return


# ============================================================= #
# ============================================================= #





#! /usr/bin/python3.6
# -*- coding: utf-8 -*-

"""
A-team coordinates in RA DEC
        by A. Loh
"""

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['Source']


from astropy import coordinates as coord

ATEAM = {'vir a': (187.70593075, +12.39112331),
         'cyg a': (299.86815263, +40.73391583),
         'cas a': (350.850000, +58.815000),
         'her a': (252.783433, +04.993031),
         'hyd a': (139.523546, -12.095553),
         'tau a': (83.63308, +22.01450)}


class Source(object):
    def __init__(self, source, time=None, location=None):
        self.time = time
        self.location = location
        self.source = source

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def source(self):
        return self._source
    @source.setter
    def source(self, s):
        # try:
        if not isinstance(s, coord.SkyCoord):
            if s.lower() in ATEAM.keys():
                src = coord.SkyCoord(ATEAM[s.lower()][0], ATEAM[s.lower()][1], frame="icrs", unit="deg")
            elif s.lower() in ['sun', 'moon', 'jupiter', 'saturn', 'mars', 'venus']:
                try:
                    with coord.solar_system_ephemeris.set('builtin'):
                        src = coord.get_body(s, self.time, self.location)
                except:
                    src = None
            else:
                src = coord.SkyCoord.from_name(s)
        else:
            src = s
        self._source = src
        # except:
        #     self._source = None

    # ================================================================= #
    # ========================= Class Methods ========================= #
    # @classmethod
    # def VirA(cls):
    #     # return cls(source='vira')
    #     self.source = coord.SkyCoord(vira[0], vira[1], frame="icrs", unit="deg")
    #     return

    # # @classmethod
    # def CygA(cls):
    #     # return cls(source='cyga')
    #     self.source = coord.SkyCoord(cyga[0], cyga[1], frame="icrs", unit="deg")
    #     return

    # # @classmethod
    # def CasA(cls):
    #     # return cls(source='casa')
    #     self.source = coord.SkyCoord(casa[0], casa[1], frame="icrs", unit="deg")
    #     return

    # # @classmethod
    # def HerA(cls):
    #     # return cls(source='hera')
    #     self.source = coord.SkyCoord(hera[0], hera[1], frame="icrs", unit="deg")
    #     return

    # # @classmethod
    # def HydA(cls):
    #     # return cls(source='hyda')
    #     self.source = coord.SkyCoord(hyda[0], hyda[1], frame="icrs", unit="deg")
    #     return





#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to compute a BST NenuFAR beam
        by A. Loh
"""

import os
import sys
import numpy as np

import matplotlib as mpl
from matplotlib import pyplot as plt
import matplotlib.ticker as mtick
from astropy.io import fits

import healpy as hp
from astropy import constants as const
from . import SSTbeamHPX

from . import SSTbeam
from . import PhasedArrayBeam
from .antenna import AntennaModel
from .antenna import miniarrays
from ..read import BST
from ..utils import ProgressBar

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['BSTbeam', 'BSTbeamHPX']

class BSTbeam(object):
    """ Class to handle BST beams
    """
    def __init__(self, bst=None, **kwargs):
        self.bst = bst
        self._evalkwargs(kwargs)

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def bst(self):
        """ BST observation
        """
        return self._bst
    @bst.setter
    def bst(self, b):
        if isinstance(b, BST):
            self._bst       = b
            self.ma         = self._bst.ma
            self.marotation = self._bst.marotation
            self.maposition = self._bst.maposition
            self.delays     = self._bst.delays
            self.freq       = self._bst.freq
            self.polar      = self._bst.polar
            self.azana      = self._bst.azana 
            self.elana      = self._bst.elana
            self.azdig      = self._bst.azdig 
            self.eldig      = self._bst.eldig
        else:
            self._bst = b
        return

    @property
    def freq(self):
        """ Frequency selection in MHz
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if not isinstance(f, (int, float, np.float32, np.float64)):
            raise TypeError("\n\t=== Attribute 'freq' should be a number ===")
        elif (f<=0.) or (f>100.):
            raise ValueError("\n\t=== Attribute 'freq' set to a value outside of NenuFAR's frequency range ===")
        else:
            self._freq = f
            return

    @property
    def polar(self):
        """ Polarization selection
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, (str)):
            raise TypeError("\n\t=== Attribute 'polar' should be a string ===")
        elif p.upper() not in ['NW', 'NE']:
            raise ValueError("\n\t=== Attribute 'polar' does not correspond to either 'NW' or 'NE' ===")
        else:
            self._polar = p.upper()
            return

    @property
    def ma(self):
        """ Miniarrays selection
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        if self.bst is None:
            marecorded = miniarrays.ma[:, 0].astype(int)
        else:
            marecorded = self.bst.ma
        if m is None:
            # print("\n\t==== WARNING: miniarrays are set by default ===")
            m = marecorded
        if isinstance(m, list):
            m = np.array(m)
        if not isinstance(m, np.ndarray):
            raise TypeError("\n\t=== Attribute 'ma' should be an array ===")
        elif not all([ mi in marecorded for mi in m]):
            raise ValueError("\n\t=== Attribute 'ma' contains miniarray indices not matching existing MAs (up to {}) ===".format(marecorded.max()))
        else:
            self._ma = m #np.unique(m)
            return

    @property
    def marotation(self):
        """ MA rotation selection
        """
        if self.bst is None:
            return self._marotation[self.ma]
        else:
            return self._marotation
    @marotation.setter
    def marotation(self, r):
        if r is None:
            # print("\n\t==== WARNING: MA rotations are set by default ===")
            self._marotation = miniarrays.ma[:, 1]
        if isinstance(r, list) or (isinstance(r, np.ndarray)):
            self._marotation = np.array(r)
            if self._marotation.size != self.ma.size:
                raise ValueError("\n\t=== Attribute 'marotation' should be the same size as 'ma' ===")
        return

    @property
    def maposition(self):
        """ MA position selection
        """
        if self.bst is None:
            return self._maposition[self.ma]
        else:
            return self._maposition
    @maposition.setter
    def maposition(self, p):
        if p is None:
            # print("\n\t==== WARNING: MA positions are set by default ===")
            self._maposition = miniarrays.ma[:, 2:5]
        if isinstance(p, list) or (isinstance(p, np.ndarray)):
            self._maposition = np.array(p)
            if self._maposition.shape[0] != self.ma.size:
                raise ValueError("\n\t=== Attribute 'maposition' should be the same size as 'ma' ===")
        return

    @property
    def delays(self):
        """ MA delay selection (in ns)
        """
        if self.bst is None:
            return self._delays[self.ma]
        else:
            return self._delays
    @delays.setter
    def delays(self, d):
        if d is None:
            # print("\n\t==== WARNING: MA delays are set by default ===")
            self._delays = miniarrays.ma[:, 5]
        if isinstance(d, list) or isinstance(d, np.ndarray):
            self._delays = np.array(d)
            if self._delays.size != self.ma.size:
                raise ValueError("\n\t=== Attribute 'delays' should be the same size as 'ma' ===")
        return

    @property
    def azana(self):
        """ Azimuth during the analogic pointing self.abeam
        """
        return self._azana
    @azana.setter
    def azana(self, a):
        if isinstance(a, (list, np.ndarray)):
            a = a[0]
        self._azana = a
        return

    @property
    def elana(self):
        """ Elevation during the analogic pointing self.abeam
        """
        return self._elana
    @elana.setter
    def elana(self, e):
        if isinstance(e, (list, np.ndarray)):
            e = e[0]
        self._elana = e
        return

    @property
    def azdig(self):
        """ Azimuth during the numeric pointing self.dbeam
        """
        return self._azdig
    @azdig.setter
    def azdig(self, a):
        if isinstance(a, (list, np.ndarray)):
            a = a[0]
        self._azdig = a
        return

    @property
    def eldig(self):
        """ Elevation during the numeric pointing self.dbeam
        """
        return self._eldig
    @eldig.setter
    def eldig(self, e):
        if isinstance(e, (list, np.ndarray)):
            e = e[0]
        self._eldig = e
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getBeam(self, **kwargs):
        """ Get the BST beam
        """
        # from astropy.time import Time
        ma_beams = []
        # bar = ProgressBar(valmax=self.ma.size, title='Computing BST beam')
        beams = {}
        for i in range(self.ma.size):
            if str(self.marotation[i]%60) not in beams.keys():
                # Only have to compute it 6 times
                mabeam = SSTbeam(ma=self.ma[i], freq=self.freq, polar=self.polar,
                    azana=self.azana, elana=self.elana, marotation=self.marotation[i])
                mabeam.getBeam()
                ma_model = AntennaModel( design=mabeam.beam, freq=self.freq)
                # bar.update()
                beams[str(self.marotation[i]%60)] = ma_model
            # ma_beams.append( ma_model )
            ma_beams.append( beams[str(self.marotation[i]%60)] )
        beam = PhasedArrayBeam(p=self.maposition, m=ma_beams, a=self.azdig, e=self.eldig)
        self.beam = beam.getBeam()

    def plot(self, store=None, **kwargs):
        """ Plot the BST Beam
        """
        if not hasattr(self, 'beam'):
            self.getBeam(**kwargs)

        theta = np.linspace(0., 90., self.beam.shape[0])
        phi   = np.radians( np.linspace(0., 360., self.beam.shape[1]) )
        # ------ Plot ------ #
        fig = plt.figure(figsize=(18/2.54, 18/2.54))
        ax  = fig.add_subplot(111, projection='polar')
        normcb = mpl.colors.LogNorm(vmin=self.beam.max() * 1.e-4, vmax=self.beam.max())
        p = ax.pcolormesh(phi, theta, self.beam, norm=normcb, rasterized=True, **kwargs)
        ax.grid(linestyle='-', linewidth=0.5, color='white', alpha=0.4)
        plt.setp(ax.get_yticklabels(), rotation='horizontal', color='white')
        g = lambda x,y: r'%d'%(90-x)
        ax.yaxis.set_major_formatter(mtick.FuncFormatter( g ))
        if store is not None:
            plt.savefig(store, dpi=300, facecolor='none',
                edgecolor='none', transparent=True, bbox_inches='tight')
        else:
            plt.title('pol={}, freq={:.2f}MHz, az={}, el={}'.format(self.polar, self.freq, self.azana, self.elana))
            plt.show()
        # ax.axes.get_yaxis().set_ticks([])
        # plt.savefig('/Users/aloh/Desktop/natacha.pdf', dpi=200)
        plt.close('all')
        return

    def save(self, savefile=None):
        """ Save the beam
        """
        if savefile is None:
            savefile = 'beam.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        if not hasattr(self, 'beam'):
            self.getBeam()

        prihdr = fits.Header()
        prihdr.set('FREQ', str(self.freq))
        prihdr.set('POLAR', self.polar)
        # prihdr.set('MINI-ARR', str(self.ma))
        # prihdr.set('MA-ROT', str(self.rotation))
        datahdu = fits.PrimaryHDU( np.fliplr(self.beam).T, header=prihdr)
        hdulist = fits.HDUList([datahdu])
        hdulist.writeto(savefile, overwrite=True)
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _evalkwargs(self, kwargs):
        """
        Parameters
        ----------
        kwargs : dictionnary
            Dictionnary of keys and values to look from in order to fill the class attributes
        """
        allowed = ['ma', 'marotation', 'maposition', 'delays', 'freq', 'polar', 'azana', 'elana', 'azdig', 'eldig']
        if not all([ki in allowed for ki in kwargs.keys()]):
            print("\n\t=== WARNING: unkwnown keywords, authorized: {} ===".format(allowed))
        if not hasattr(self, 'ma'):
            self.ma         = kwargs.get('ma', None)
        if not hasattr(self, 'marotation'):
            self.marotation = kwargs.get('marotation', None)
        if not hasattr(self, 'maposition'):
            self.maposition = kwargs.get('maposition', None)
        if not hasattr(self, 'delays'):
            self.delays     = kwargs.get('delays', None)
        if not hasattr(self, 'freq'):
            self.freq       = kwargs.get('freq', 50)
        if not hasattr(self, 'polar'):
            self.polar      = kwargs.get('polar', 'NW')
        if not hasattr(self, 'azana'):
            self.azana      = kwargs.get('azana', 180.)
        if not hasattr(self, 'elana'):
            self.elana      = kwargs.get('elana', 90.)
        if not hasattr(self, 'azdig'):
            self.azdig      = kwargs.get('azdig', 180.)
        if not hasattr(self, 'eldig'):
            self.eldig      = kwargs.get('eldig', 90.)
        return















# ============================================================= #
# ============================================================= #
class BSTbeamHPX(object):
    """ Class to handle BST beams
    """
    def __init__(self, bst=None, **kwargs):
        self.bst = bst
        self._evalkwargs(kwargs)

    # ========================================================= #
    # ==================== Getter / Setter ==================== #
    @property
    def bst(self):
        """ BST observation
        """
        return self._bst
    @bst.setter
    def bst(self, b):
        if isinstance(b, BST):
            self._bst       = b
            self.ma         = self._bst.ma
            self.marotation = self._bst.marotation
            self.maposition = self._bst.maposition
            self.delays     = self._bst.delays
            self.freq       = self._bst.freq
            self.polar      = self._bst.polar
            self.azana      = self._bst.azana 
            self.elana      = self._bst.elana
            self.azdig      = self._bst.azdig 
            self.eldig      = self._bst.eldig
        else:
            self._bst = b
        return

    # --------------------------------------------------------- #
    @property
    def freq(self):
        """ Frequency selection in MHz
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if not isinstance(f, (int, float, np.float32, np.float64)):
            raise TypeError("\n\t=== Attribute 'freq' should be a number ===")
        elif (f<=0.) or (f>100.):
            raise ValueError("\n\t=== Attribute 'freq' set to a value outside of NenuFAR's frequency range ===")
        else:
            self._freq = f
            self.wavevec = 2 * np.pi * f * 1.e6 / const.c.value
            return

    # --------------------------------------------------------- #
    @property
    def polar(self):
        """ Polarization selection
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, (str)):
            raise TypeError("\n\t=== Attribute 'polar' should be a string ===")
        elif p.upper() not in ['NW', 'NE']:
            raise ValueError("\n\t=== Attribute 'polar' does not correspond to either 'NW' or 'NE' ===")
        else:
            self._polar = p.upper()
            return

    # --------------------------------------------------------- #
    @property
    def ma(self):
        """ Miniarrays selection
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        if self.bst is None:
            marecorded = miniarrays.ma[:, 0].astype(int)
        else:
            marecorded = self.bst.ma
        if m is None:
            # print("\n\t==== WARNING: miniarrays are set by default ===")
            m = marecorded
        if isinstance(m, list):
            m = np.array(m)
        if not isinstance(m, np.ndarray):
            raise TypeError("\n\t=== Attribute 'ma' should be an array ===")
        elif not all([ mi in marecorded for mi in m]):
            raise ValueError("\n\t=== Attribute 'ma' contains miniarray indices not matching existing MAs (up to {}) ===".format(marecorded.max()))
        else:
            self._ma = m #np.unique(m)
            return

    # --------------------------------------------------------- #
    @property
    def marotation(self):
        """ MA rotation selection
        """
        if self.bst is None:
            return self._marotation[self.ma]
        else:
            return self._marotation
    @marotation.setter
    def marotation(self, r):
        if r is None:
            # print("\n\t==== WARNING: MA rotations are set by default ===")
            self._marotation = miniarrays.ma[:, 1]
        if isinstance(r, list) or (isinstance(r, np.ndarray)):
            self._marotation = np.array(r)
            if self._marotation.size != self.ma.size:
                raise ValueError("\n\t=== Attribute 'marotation' should be the same size as 'ma' ===")
        return

    # --------------------------------------------------------- #
    @property
    def maposition(self):
        """ MA position selection
        """
        if self.bst is None:
            return self._maposition[self.ma]
        else:
            return self._maposition
    @maposition.setter
    def maposition(self, p):
        if p is None:
            # print("\n\t==== WARNING: MA positions are set by default ===")
            self._maposition = miniarrays.ma[:, 2:5]
        if isinstance(p, list) or (isinstance(p, np.ndarray)):
            self._maposition = np.array(p)
            if self._maposition.shape[0] != self.ma.size:
                raise ValueError("\n\t=== Attribute 'maposition' should be the same size as 'ma' ===")
        return

    # --------------------------------------------------------- #
    @property
    def delays(self):
        """ MA delay selection (in ns)
        """
        if self.bst is None:
            return self._delays[self.ma]
        else:
            return self._delays
    @delays.setter
    def delays(self, d):
        if d is None:
            # print("\n\t==== WARNING: MA delays are set by default ===")
            self._delays = miniarrays.ma[:, 5]
        if isinstance(d, list) or isinstance(d, np.ndarray):
            self._delays = np.array(d)
            if self._delays.size != self.ma.size:
                raise ValueError("\n\t=== Attribute 'delays' should be the same size as 'ma' ===")
        return

    # --------------------------------------------------------- #
    @property
    def azana(self):
        """ Azimuth during the analogic pointing self.abeam
        """
        return self._azana
    @azana.setter
    def azana(self, a):
        if isinstance(a, (list, np.ndarray)):
            a = a[0]
        self._azana = a
        return

    # --------------------------------------------------------- #
    @property
    def elana(self):
        """ Elevation during the analogic pointing self.abeam
        """
        return self._elana
    @elana.setter
    def elana(self, e):
        if isinstance(e, (list, np.ndarray)):
            e = e[0]
        self._elana = e
        return

    # --------------------------------------------------------- #
    @property
    def azdig(self):
        """ Azimuth during the numeric pointing self.dbeam
        """
        return self._azdig
    @azdig.setter
    def azdig(self, a):
        if isinstance(a, (list, np.ndarray)):
            a = a[0]
        self._azdig = np.radians(a)
        return

    # --------------------------------------------------------- #
    @property
    def eldig(self):
        """ Elevation during the numeric pointing self.dbeam
        """
        return self._eldig
    @eldig.setter
    def eldig(self, e):
        if isinstance(e, (list, np.ndarray)):
            e = e[0]
        self._eldig = np.radians(e)
        return

    # --------------------------------------------------------- #
    @property
    def resolution(self):
        """ Angular resolution in degrees
        """
        return self._resolution
    @resolution.setter
    def resolution(self, r):
        # Find the nside corresponding to the desired resolution
        nsides = np.array([2**i for i in range(1, 12)])
        resol_deg = np.degrees(hp.nside2resol(nside=nsides, arcmin=False))
        idx = (np.abs(resol_deg - r)).argmin()
        self._nside = nsides[idx]
        self._npix = hp.nside2npix(self._nside)        
        self._resolution = resol_deg[idx]

        self._sky_grid()
        return


    # ========================================================= #
    def compute(self):
        """ Compute the phased array beam
        """
        ma_beams = []
        beams = {}
        for i in range(self.ma.size):
            if str(self.marotation[i]%60) not in beams.keys():
                
                mabeam = SSTbeamHPX(ma=self.ma[i], freq=self.freq, polar=self.polar,
                    azana=self.azana, elana=self.elana, marotation=self.marotation[i],
                    resolution=self.resolution)
                mabeam.compute()
                beams[str(self.marotation[i]%60)] = mabeam.beam
            ma_beams.append( beams[str(self.marotation[i]%60)] )
        ma_beams = np.array(ma_beams)
       
        # ------ Phase delay towards pointing ------ #
        ux = np.cos(self.azdig) * np.cos(self.eldig)
        uy = np.sin(self.azdig) * np.cos(self.eldig)
        uz = np.cos(self.eldig)
        phix = self.maposition[:, 0] * ux[np.newaxis]
        phiy = self.maposition[:, 1] * uy[np.newaxis]
        phiz = self.maposition[:, 2] * uz[np.newaxis]
        phi  = self.wavevec * (phix + phiy + phiz)

        # ------ Phase reference ------ #
        _xp = np.cos(self._az) * np.cos(self._alt) 
        _yp = np.sin(self._az) * np.cos(self._alt) 
        _zp = np.sin(self._alt)
        phi0x = self.maposition[:, 0] * _xp[:, np.newaxis]
        phi0y = self.maposition[:, 1] * _yp[:, np.newaxis]
        phi0z = self.maposition[:, 2] * _zp[:, np.newaxis]
        phi0  = self.wavevec * (phi0x + phi0y + phi0z)
        
        phase = phi0[:, :] - phi[np.newaxis, :]

        # ------ e^(i Phi) ------ #
        eiphi = np.sum( np.exp(1j * phase), axis=1 )
        self.beam = np.real(eiphi * eiphi.conjugate()) * np.sum(ma_beams, axis=0)
        self.beam[self.below_horizon] = 0.

        return

    # --------------------------------------------------------- #
    def plot(self):
        """
        """
        if not hasattr(self, 'beam'):
            self.compute()
        # hp.orthview( np.log10(self.beam), rot=[0, 90, 90])
        # hp.graticule()
        # plt.show()
        # plt.close('all')

        hp.cartview( np.log10(self.beam))
        # hp.graticule()
        plt.show()
        plt.close('all')
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _sky_grid(self):
        """ Compute the sky grid
        """
        self._az, self._alt = np.radians(hp.pix2ang(nside=self._nside,
                                            ipix=np.arange(self._npix),
                                            lonlat=True,
                                            nest=False))
        self._az = self._az[::-1] # east is 90        
        # Remove local altitudes < 0deg
        self.below_horizon = (self._alt < 0.)
        return

    # --------------------------------------------------------- #
    def _evalkwargs(self, kwargs):
        """
        Parameters
        ----------
        kwargs : dictionnary
            Dictionnary of keys and values to look from in order to fill the class attributes
        """
        allowed = ['ma', 'marotation', 'maposition', 'delays', 'freq', 'polar', 'azana', 'elana', 'azdig', 'eldig', 'resolution']
        if not all([ki in allowed for ki in kwargs.keys()]):
            print("\n\t=== WARNING: unkwnown keywords, authorized: {} ===".format(allowed))
        if not hasattr(self, 'ma'):
            self.ma         = kwargs.get('ma', None)
        if not hasattr(self, 'marotation'):
            self.marotation = kwargs.get('marotation', None)
        if not hasattr(self, 'maposition'):
            self.maposition = kwargs.get('maposition', None)
        if not hasattr(self, 'delays'):
            self.delays     = kwargs.get('delays', None)
        if not hasattr(self, 'freq'):
            self.freq       = kwargs.get('freq', 50)
        if not hasattr(self, 'polar'):
            self.polar      = kwargs.get('polar', 'NW')
        if not hasattr(self, 'azana'):
            self.azana      = kwargs.get('azana', 180.)
        if not hasattr(self, 'elana'):
            self.elana      = kwargs.get('elana', 90.)
        if not hasattr(self, 'azdig'):
            self.azdig      = kwargs.get('azdig', 180.)
        if not hasattr(self, 'eldig'):
            self.eldig      = kwargs.get('eldig', 90.)
        if not hasattr(self, 'resolution'):
            self.resolution = kwargs.get('resolution', 0.2)
        return






#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to read NenuFAR SST data
        by A. Loh
"""

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.1.0'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'Production'
__all__ = ['SST']


import os
import sys
import glob
import numpy as np

from matplotlib.colors import LogNorm
from matplotlib import pyplot as plt

from astropy.io import fits
from astropy.time import Time, TimeDelta


class SST(object):
    def __init__(self, obsfile=None):
        self.obsfile = obsfile
        self.ma      = 0
        self.freq    = 50
        self.polar   = 'nw'
        self.abeam   = 0 

        self._attrlist = ['polar', 'freq', 'time', 'abeam', 'ma']

    def __str__(self):
        toprint  = '\t=== Class SST of nenupy ===\n'
        toprint += '\tList of all current attributes:\n'
        for att in dir(self):
            avoid = ['data', 'elana', 'azana', 'maposition', 'marotation', 'plot', 'save', 'select']
            if (not att.startswith('_')) & (not any(x.isupper() for x in att)) & (att not in avoid):
                toprint += "%s: %s\n"%(att, getattr(self, att))
        return toprint

    def __repr__(self):
        return '<SST object: obsfile={}>'.format(self.obsfile)

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def obsfile(self):
        """ SST observation file,
            obsfile set to 'fake' will produce a custom fake SST observation
        """
        return self._obsfile
    @obsfile.setter
    def obsfile(self, o):
        if o == 'fake':
            self._createFakeObs()
            self._obsfile = o
            return
        
        if (o is None) or (o == '') or (isinstance(o, str)) and (not os.path.isfile(o)):
            # Look at the current/specified directory
            if o is None:
                o = ''
            if os.path.isdir(o):
                _opath = os.path.abspath(o)
                sstfiles = glob.glob( os.path.join(_opath, '*SST.fits') )
            else:
                sstfiles = glob.glob('*SST.fits')
            if len(sstfiles) == 0:
                raise IOError("\t=== No SST fits file in current directory, specify the file to read. ===")
            else:
                o = [os.path.abspath(f) for f in sstfiles]  
        elif isinstance(o, list):
            # List of files in input
            if all(os.path.isfile(f) for f in o):
                if all(f.endswith('SST.fits') for f in o):
                    o = [os.path.abspath(f) for f in o]
                else:
                    raise IOError("\t=== Multiple observation files (other than SST) are not understood - yet. ===")
            else:
                raise IOError("\t=== At least one file not found. ===")
        else:
            if not os.path.isfile(o):
                raise IOError("\t=== File {} not found. ===".format(os.path.abspath(o)))
            else:
                # The file has been correctly found
                o = [ os.path.abspath(o) ]
        self._obsfile = sorted(o)

        if not self._isSST():
            raise ValueError("\t=== Files might not all be SST observations ===")
        else:
            self._readSST()

        return

    @property
    def ma(self):
        """ Selected Mini-Array
        """
        return self._ma
    @ma.setter
    def ma(self, m):
        if m is None:
            m = 0
        if not isinstance(m, int):
            try:
                m = int(m)
            except:
                print("\n\t=== WARNING: Mini-Array index {} not recognized ===".format(m))
                m = 0
        if m in self.miniarrays:
                self._ma = m
        else:
            print("\n\t=== WARNING: available Mini-Arrays are {} ===".format(self.miniarrays))
            self._ma = 0
        return

    @property
    def freq(self):
        """ Frequency selection in MHz
            Example:
            self.freq = 50
            self.freq = [20, 30]
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if isinstance(f, list):
            self._freq    = [0, 0]
            self._freq[0] = min(self._freqs, key=lambda x: np.abs(x-f[0]))
            self._freq[1] = min(self._freqs, key=lambda x: np.abs(x-f[1]))
        else:
            self._freq    = min(self._freqs, key=lambda x: np.abs(x-f))  
        return        

    @property
    def polar(self):
        """ Polarization selection ('NW' or 'NE')
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, str):
            self._polar = 'NW'
        elif p.upper() not in self._pols:
            print("\n\t=== WARNING: Polarization selection {} not recognized ===".format(p))
            self._polar = 'NW'
        else:
            self._polar = p.upper()

    # @property
    # def time(self):
    #     """ Time selection
    #     """
    #     return self._time
    # @time.setter
    # def time(self, t):
    #     if isinstance(t, list):
    #         try:
    #             if not isinstance(t[0], Time):
    #                 t[0] = Time(t[0])
    #             if not isinstance(t[1], Time):
    #                 t[1] = Time(t[1])
    #             self._time = t
    #         except:
    #             print("\n\t=== WARNING: Time syntax incorrect ===")
    #     else:
    #         if not isinstance(t, Time):
    #             try:
    #                 t = Time(t)
    #             except:
    #                 print("\n\t=== WARNING: Time syntax incorrect ===")
    #         self._time = t
    #         return

    @property
    def time(self):
        """ Time selection
        """
        return self._time
    @time.setter
    def time(self, t):
        if isinstance(t, list):
            if not isinstance(t[0], Time):
                t[0] = Time(t[0])
                
            if not isinstance(t[1], Time):
                t[1] = Time(t[1])
            assert self.obstart <= t[0] <= self.obstop, 'time[0] outside time range'
            assert self.obstart <= t[1] <= self.obstop, 'time[1] outside time range'
            assert t[0] < t[1], 'time[0] > time[1]!'
            self._time = t
        else:
            if not isinstance(t, Time):
                t = Time(t)
            assert self.obstart <= t <= self.obstop, 'time outside time range'
            self._time = t
            return

    @property
    def abeam(self):
        """ Index of the selected analogic beam
        """
        return self._abeam
    @abeam.setter
    def abeam(self, a):
        if (a is None) or (a==-1):
            return
        elif a not in self.abeams:
            print("\n\t=== WARNING: available AnaBeams are {} ===".format(self.abeams))
            self._abeam = 0
        else:
            self._abeam = a
        return

    # ------ Specific getters ------ #
    @property
    def marotation(self):
        """ Mini-array rotation in degrees
        """
        self._marotation = self._marot[ self.miniarrays == self.ma ][0]
        return self._marotation

    @property
    def maposition(self):
        """ Lambert93 (x, y, z) Mini-Array position
        """
        self._maposition = self._mapos[ self.miniarrays == self.ma ][0]
        return self._maposition
    
    @property
    def azana(self):
        """ Azimuth during the analogic pointing self.abeam
        """
        az = self._azlistana[ self._pointana==self.abeam ]
        if az.size == 1:
            az = az[0]
        else:
            anatimes = self._pointanat[ self._pointana==self.abeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((anatimes >= self.time[0]) & (anatimes <= self.time[1]))
                az = az[tmask]
            else:
                # az = az[ np.argmin(np.abs( (anatimes - self.time).sec )) ]
                az = az[(anatimes <= self.time)][-1]
        return az 

    @property
    def elana(self):
        """ Elevation during the analogic pointing self.abeam
        """
        el = self._ellistana[ self._pointana==self.abeam ]
        if el.size == 1:
            el = el[0]
        else:
            anatimes = self._pointanat[ self._pointana==self.abeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((anatimes >= self.time[0]) & (anatimes <= self.time[1]))
                # el = self._ellistana[tmask]
                el = el[tmask]
            else:
                # el = self._ellistana[ np.argmin(np.abs( (anatimes - self.time).sec )) ]
                el = el[(anatimes <= self.time)][-1]
        return el

    # ================================================================= #
    # =========================== Methods ============================= #
    def select(self, **kwargs):
        """ Make the data selection

            Parameters
            ----------
            kwargs : {freq, ma, polar, time}
                Keyword arguments

            Returns
            -------
            self.d : np.ndarray
                Data selected
            self.f : np.ndarray
                Frequency selected
            self.t : np.ndarray
                Time selected
        """
        self._evalkwargs(kwargs)
        for ssfile in sorted(self.obsfile):
            with fits.open(ssfile, mode='readonly', ignore_missing_end=True, memmap=True) as fitsfile:
                all_time = fitsfile[7].data['jd']
                # ------ polarization ------ #
                mask_pol = (self._pols == self.polar)
                # ------ frequency ------ #
                if isinstance(self.freq, list):
                    mask_fre = (self._freqs > 0.) & (self._freqs >= self.freq[0]) & (self._freqs <= self.freq[-1])
                else:
                    mask_fre = (self._freqs == min(self._freqs, key=lambda x: np.abs(x-self.freq)))
                # ------ time ------ #
                time_obj = Time(all_time, format='jd')
                if isinstance(self.time, list):
                    mask_tim = (time_obj >= self.time[0]) & (time_obj <= self.time[1])
                else:
                    mask_tim = (time_obj.mjd == min(time_obj.mjd, key=lambda x: np.abs(x-self.time.mjd)) )
                # ------ mini-array ------ #
                mask_ma = (self.miniarrays == self.ma)
                if not 'this_data' in locals():
                    this_data = fitsfile[7].data['data'][np.ix_(mask_tim, mask_ma, mask_pol, mask_fre)]
                    this_time = all_time[mask_tim]
                else:
                    this_data = np.vstack( (this_data, fitsfile[7].data['data'][np.ix_(mask_tim, mask_ma, mask_pol, mask_fre)]))
                    this_time = np.hstack( (this_time, all_time[mask_tim]) )
        f = self._freqs[ mask_fre ]
        self.data = {'amp': np.squeeze(this_data), 'freq': f, 'time': Time(np.squeeze(this_time), format='jd')}
        return
    def select_v0(self, **kwargs):
        """ Make the data selection

            Parameters
            ----------
            kwargs : {freq, ma, polar, time}
                Keyword arguments

            Returns
            -------
            self.d : np.ndarray
                Data selected
            self.f : np.ndarray
                Frequency selected
            self.t : np.ndarray
                Time selected
        """
        self._evalkwargs(kwargs)

        # ------ load data only once ------ #
        if not hasattr(self, '_dataall'):
            self._timeall = []
            self._dataall = []
            for ssfile in sorted(self.obsfile):
                with fits.open(ssfile, mode='readonly', ignore_missing_end=True, memmap=True) as fitsfile:
                   self._timeall.append( fitsfile[7].data['jd']   ) 
                   self._dataall.append( fitsfile[7].data['data'] )
            self._timeall = Time( np.hstack(self._timeall), format='jd' )
            self._dataall  = np.vstack( self._dataall )

        # ------ polarization ------ #
        mask_pol = (self._pols == self.polar)

        # ------ frequency ------ #
        if isinstance(self.freq, list):
            mask_fre = (self._freqs > 0.) & (self._freqs >= self.freq[0]) & (self._freqs <= self.freq[-1])
        else:
            mask_fre = (self._freqs == min(self._freqs, key=lambda x: np.abs(x-self.freq)))

        # ------ time ------ #
        if isinstance(self.time, list):
            mask_tim = (self._timeall >= self.time[0]) & (self._timeall <= self.time[1])
        else:
            mask_tim = (self._timeall.mjd == min(self._timeall.mjd, key=lambda x: np.abs(x-self.time.mjd)) )
        
        # ------ mini-array ------ #
        mask_ma = (self.miniarrays == self.ma)
       
        # ------ selected data ------ #
        # self.d = np.squeeze( self._dataall[ np.ix_(mask_tim, mask_ma, mask_pol, mask_fre) ] )
        # self.t = self._timeall[ mask_tim ]
        # self.f = self._freqs[ mask_fre ]
        d = np.squeeze( self._dataall[ np.ix_(mask_tim, mask_ma, mask_pol, mask_fre) ] )
        t = self._timeall[ mask_tim ]
        f = self._freqs[ mask_fre ]
        self.data = {'amp': d, 'freq': f, 'time': t}
        return

    def plot(self, db=True, **kwargs):
        """ Plot the data
        """
        self.select(**kwargs)

        plotkwargs = {key: value for (key, value) in kwargs.items() if key not in self._attrlist}

        if self.data['freq'].size == 1:
            # ------ Light curve ------ #
            xtime = (self.data['time'] - self.data['time'][0]).sec / 60
            if db:
                plt.plot(xtime, 10.*np.log10(self.data['amp']), **plotkwargs)
                plt.ylabel('dB')
            else:
                plt.plot(xtime, self.data['amp'], **plotkwargs)
                plt.ylabel('Amplitude')
            plt.xlabel('Time (min since {})'.format(self.data['time'][0].iso))
            plt.title('MA={}, f={:3.2f} MHz, pol={}'.format(self.ma, self.freq, self.polar))

        elif self.data['time'].size == 1:
            # ------ Spectrum ------ #
            if db:
                plt.plot(self.data['freq'], 10.*np.log10(self.data['amp']), **plotkwargs)
                plt.ylabel('dB')
            else:
                plt.plot(self.data['freq'], self.data['amp'], **plotkwargs)
                plt.ylabel('Amplitude')
            plt.xlabel('Frequency (MHz)')
            plt.title('MA={}, t={}, pol={}'.format(self.ma, self.time.iso, self.polar))

        elif (self.data['freq'].size > 1) & (self.data['time'].size > 1):
            # ------ Dynamic spectrum ------ #
            xtime = (self.data['time'] - self.data['time'][0]).sec / 60
            vmin, vmax = np.percentile(self.data['amp'], [5, 99]) 
            cmap = 'bone'
            for key, value in plotkwargs.items():
                if key == 'cmap': cmap = value
                if key == 'vmin': vmax = value
                if key == 'vmax': vmin = value
            fig = plt.figure()
            ax  = fig.add_subplot(111)
            normcb = LogNorm(vmin=vmin, vmax=vmax)
            spec   = ax.pcolormesh(xtime, self.data['freq'], self.data['amp'].T, cmap=cmap, norm=normcb)
            plt.colorbar(spec)
            ax.axis( [xtime.min(), xtime.max(), self.data['freq'].min(), self.data['freq'].max()] )
            plt.xlabel('Time (min since {})'.format(self.data['time'][0].iso))
            plt.ylabel('Frequency (MHz)')
            plt.title('MA={}, pol={}'.format(self.ma, self.polar))

        else:
            raise ValueError("\n\t=== ERROR: Plot nature not understood ===")
        plt.show()
        plt.close('all')
        return

    def save(self, savefile=None, **kwargs):
        """ Save the data
        """
        if savefile is None:
            savefile = self.obsname + '_data.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        self.select(**kwargs)

        prihdr = fits.Header()
        prihdr.set('OBS', self.obsname)
        prihdr.set('FREQ', str(self.freq))
        prihdr.set('TIME', str(self.time))
        prihdr.set('POLAR', self.polar)
        prihdr.set('MINI-ARR', str(self.ma))
        datahdu = fits.PrimaryHDU(self.d.T, header=prihdr)
        freqhdu = fits.BinTableHDU.from_columns( [fits.Column(name='frequency', format='D', array=self.f)] )
        timehdu = fits.BinTableHDU.from_columns( [fits.Column(name='mjd', format='D', array=self.t.mjd)] )
        hdulist = fits.HDUList([datahdu, freqhdu, timehdu])
        hdulist.writeto(savefile, overwrite=True)
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _isSST(self):
        """ Check that self.obsfile contain only SST observations
        """
        isSST = True
        for sstfile in self.obsfile:
            with fits.open(sstfile, mode='readonly', ignore_missing_end=True, memmap=True) as f:
                if not 'OBJECT' in f[0].header.keys():
                    # This is an old fits
                    break
                if f[0].header['OBJECT'].lower() != 'subband statistics':
                    isSST = False
                else:
                    pass
        return isSST

    def _readSST(self):
        """ Read SST fits files and fill the class attributes
        """
        self.obsname = os.path.basename( self.obsfile[0] )[:-9]
        
        headi = fits.getheader(self.obsfile[0], ext=0, ignore_missing_end=True, memmap=True)
        headf = fits.getheader(self.obsfile[-1], ext=0, ignore_missing_end=True, memmap=True)
        # if headi['DATE-END'] == '':
        #     headi['DATE-END'] = headi['DATE-OBS']
        # if headf['DATE-END'] == '':
        #     headf['DATE-END'] = headf['DATE-OBS']
        if not hasattr(self, 'obstart'):
            self.obstart  = Time( headi['DATE-OBS'] + 'T' + headi['TIME-OBS'] )
        self.obstop   = Time( headf['DATE-END'] + 'T' + headf['TIME-END'] )
        self.time     = [self.obstart.copy(), self.obstop.copy()]
        self.exposure = self.obstop - self.obstart

        setup_ins = fits.getdata(self.obsfile[0], ext=1, ignore_missing_end=True, memmap=True)
        self._freqs     = np.squeeze( setup_ins['frq'] )
        self.freqmin    = self._freqs.min()
        self.freqmax    = self._freqs.max()
        self.miniarrays = np.squeeze( setup_ins['noMROn'] )
        self._marot     = np.squeeze( setup_ins['rotation'] )
        self._mapos     = np.squeeze( setup_ins['noPosition'] )
        self._mapos     = self._mapos.reshape( int(self._mapos.size/3), 3 )
        self._pols      = np.squeeze( setup_ins['spol'] )

        self._readParset()

        # if self._pointanat.shape[1] != 1:
        #     self.type = 'tracking'
        # else:
        self.type = 'transit'
        return

    def _readParset(self):
        """ The SST Fits header is lacking some crucial informations
            It is thus needed to read the parset file
        """
        try:
            sstpath = os.path.dirname( self.obsfile[0] )
            parsetfile = glob.glob( os.path.join(sstpath, '{}_*.parset'.format(self.obsname.split('_')[0])) )[0]
            with open(parsetfile) as rf:
                parset = rf.read()
            parsedfile = parset.split('\n')

            self.abeams     = np.arange([int(i.split('=')[1]) for i in parsedfile if 'nrAnaBeams' in i][0])
            self._antlist   = np.array([ np.array(eval(i.split('=')[1]))-1 for i in sorted(parsedfile) if 'antList' in i ])
            self._adeltat   = TimeDelta( np.array([float(i.split('=')[1]) for i in sorted(parsedfile) if ('duration' in i) & ('AnaBeam' in i) ]), format='sec')
            self._pointana  = self.abeams.copy()
            self._azlistana = np.array([float(i.split('=')[1]) for i in sorted(parsedfile) if ('angle1' in i) & ('AnaBeam' in i) ])
            self._ellistana = np.array([float(i.split('=')[1]) for i in sorted(parsedfile) if ('angle2' in i) & ('AnaBeam' in i) ])
            self._pointanat = Time(np.array([i.split('=')[1] for i in sorted(parsedfile) if ('startTime' in i) & ('AnaBeam' in i) ]))
        except:
            # default values
            self.abeams     = np.zeros(1)
            self._antlist   = np.array([np.arange(19)])
            self._adeltat   = TimeDelta( np.array([24 * 3600.]), format='sec')
            self._pointana  = np.zeros(1)
            self._azlistana = np.array([180])
            self._ellistana = np.array([90])
            self._pointanat = Time(np.array([self.obstart.iso]))
        return

    def _createFakeObs(self):
        """
        """
        return

    def _evalkwargs(self, kwargs):
        for key, value in kwargs.items():
            if   key == 'polar': self.polar = value
            elif key == 'freq':  self.freq  = value
            elif key == 'time':  self.time  = value
            elif key == 'ma':    self.ma    = value
            elif key == 'abeam': self.abeam = value
            else:
                pass
        return




from setuptools import setup, find_packages
import re
import nenupy

meta_file = open('nenupy/metadata.py').read()
metadata  = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", meta_file))

setup(name       = 'nenupy',
    packages     = find_packages(),
    include_package_data=True,
    install_requires=['numpy', 'scipy', 'astropy', 'matplotlib'],
    python_requires='>=3.5',
    scripts      = ['bin/nenuplot', 'bin/nenusim', 'bin/nenucor', 'bin/nenuinfo', 'bin/nenuplot-gui', 'bin/nenubeam', 'bin/nenuskybeam'],
    version      = nenupy.__version__,
    description  = 'NenuFAR Python package',
    url          = 'https://github.com/AlanLoh/nenupy.git',
    author       = metadata['author'],
    author_email = metadata['email'],
    license      = 'MIT',
    zip_safe     = False)

# make the package:
# python3 setup.py sdist bdist_wheel
# upload it:
# python3 -m twine upload dist/*version*

# Release:
# git tag -a v*version* -m "annotation for this release"
# git push origin --tags

# Documentation
# sphinx-build -b html docs/ docs/_build/

# Update on nancep:
# /usr/local/bin/pip3.5 install nenupy --install-option=--prefix=/cep/lofar/nenupy3 --upgrade


#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

from .SST import SST
from .BST import BST
from .XST import XST

__all__ = ['SST', 'BST', 'XST', 'FakeObs']

class FakeObs():
    def __init__(self):
        self.freq  = 50
        self.azana = 180 
        self.elana = 90
        self.azdig = 180 
        self.eldig = 90
        self.polar = 'NW'
        # self.time  = 
